# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Sat Oct 31 12:00:00 2016

Modifications
-------------
- Sat Oct 31 12:00:00 2016

Aims
----
- APOGEE utils

"""

from __future__ import print_function

import os
import urllib
import urllib.request
from collections import OrderedDict

import numpy as np
from astropy.io import fits
from astropy.table import Table, Column

__all__ = ["apStar_url", "apStar_download", "mkdir_loop"]


def reconstruct_wcs_coord_from_fits_header(hdr, dim=1):
    """ reconstruct wcs coordinates (e.g., wavelength array) """
    # assert dim is not larger than limit
    assert dim <= hdr["NAXIS"]

    # get keywords
    crval = hdr["CRVAL%d" % dim]
    cdelt = hdr["CDELT%d" % dim]
    try:
        crpix = hdr["CRPIX%d" % dim]
    except KeyError:
        crpix = 1

    # length of the current dimension
    naxis_ = hdr["NAXIS%d" % dim]

    # reconstruct wcs coordinates
    coord = np.arange(1 - crpix, naxis_ + 1 - crpix) * cdelt + crval
    return coord


def apStar_read(fp, full=False, meta=False, verbose=False):
    """ read apStar fits file

    Parameters
    ----------
    fp: string
        file path
    full: bool
        if False, return a simple version of apStar spec.
        if True, return a full version of apStar spec.
    meta: bool
        if True, attach Primary HDU header as spec.meta (OrderedDict)
    verbose: bool:
        if True, verbose.

    Returns
    -------
    spec (astropy.table.Table instance)

    Notes
    -----
    The url of the doc for apStar files:
    https://data.sdss.org/datamodel/files/APOGEE_REDUX/APRED_VERS/APSTAR_VERS/
    TELESCOPE/LOCATION_ID/apStar.html

    HDU0: master header with target information
    HDU1: spectra: combined and individual
    HDU2: error spectra
    HDU3: mask spectra
    HDU4: sky spectra
    HDU5: sky error spectra
    HDU6: telluric spectra
    HDU7: telluric error spectra
    HDU8: table with LSF coefficients
    HDU9: table with RV/binary information

    """
    # read apStar file
    hl = fits.open(fp)

    # construct Table instance
    if not full:
        # not full apStar info, [wave, flux, flux_err, mask] only
        spec = Table([
            Column(
                10. ** reconstruct_wcs_coord_from_fits_header(hl[1].header, 1),
                "wave"),
            Column(hl[1].data.T, "flux"),
            Column(hl[2].data.T, "flux_err"),
            Column(hl[3].data.T, "mask")])
    else:
        # full apStar info
        spec = Table([
            Column(
                10. ** reconstruct_wcs_coord_from_fits_header(hl[1].header, 1),
                "wave"),
            Column(hl[1].data.T, "flux"),
            Column(hl[2].data.T, "flux_err"),
            Column(hl[3].data.T, "mask"),
            Column(hl[4].data.T, "sky"),
            Column(hl[5].data.T, "sky_err"),
            Column(hl[6].data.T, "telluric"),
            Column(hl[7].data.T, "telluric_err")])

    # meta data
    if meta:
        spec.meta = OrderedDict(hl[0].header)

    if verbose:
        print("@Cham: successfully load %s ..." % fp)
    return spec


def aspcapStar_read(fp, meta=False, verbose=False):
    """ read apStar fits file

    Parameters
    ----------
    fp: string
        file path
    meta: bool
        if True, attach Primary HDU header as spec.meta (OrderedDict)
    verbose: bool:
        if True, verbose.

    Returns
    -------
    spec (astropy.table.Table instance)

    Notes
    -----
    The url of the doc for aspcapStar files:
    https://data.sdss.org/datamodel/files/APOGEE_REDUX/APRED_VERS/APSTAR_VERS/
    ASPCAP_VERS/RESULTS_VERS/LOCATION_ID/aspcapStar.html

    HDU0: The Primary Header
    HDU1: Spectrum array
    HDU2: Error array
    HDU3: Best fit spectrum
    HDU4: ASPCAP data table

    """
    # read apStar file
    hl = fits.open(fp)

    # construct Table instance
    spec = Table([
        Column(
            10. ** reconstruct_wcs_coord_from_fits_header(hl[1].header, 1),
            "wave"),
        Column(hl[1].data, "flux"),
        Column(hl[2].data, "flux_err"),
        Column(hl[3].data, "flux_fit")])

    # meta data
    if meta:
        spec.meta = OrderedDict(hl[0].header)

    if verbose:
        print("@Cham: successfully load %s ..." % fp)
    return spec


def test_aspcapStar_read():
    fp = "/pool/sdss/apogee_dr13/aspcapStar-r6-l30e.2-2M07332578+2044059.fits"
    spec = aspcapStar_read(fp, True)
    spec.pprint()
    print(spec.meta)
    return spec
    # spec = test_aspcapStar_read()


def test_apStar_read():
    fp = "/pool/sdss/apogee_dr13/apStar-r6-VESTA.fits"
    spec = apStar_read(fp, True)
    spec.pprint()
    print(spec.meta)
    return spec
    # spec = test_apStar_read()


def apStar_url(telescope, location_id, field, file_,
               url_header=None):
    """ apStar url generator
    which in principle is able to generate file path

    Parameters
    ----------
    telescope: string
        TELESCOPE, {"apo1m', 'apo25m'}
    location_id: int
        for 'apo1m', it's 1
        for 'apo25m', it's like PLATE
    field: string
        for 'apo1m', it's 'hip'|'calibration'|...
        for 'apo25m', it's non-sense
    file_: string
        FILE
    url_header: string
        if None|'sas', it's set to be
        "https://data.sdss.org/sas/dr13/apogee/spectro/redux/%s/stars"%version

    Returns
    -------
    url: string
        the url of apStar file

    Note
    ----
    version: string
        currently it's 'r6' @20161031
    """

    if url_header is None or url_header is "sas":
        url_header = ("https://data.sdss.org/sas/dr13/apogee"
                      "/spectro/redux/r6/stars")

    url_header = url_header.strip()
    telescope = telescope.strip()
    field = field.strip()
    file_ = file_.strip()

    if telescope == "apo1m":
        # apo1m
        url = "%s/%s/%s/%s" % (url_header, telescope, field, file_)
    elif telescope == "apo25m":
        # apo25m
        url = "%s/%s/%s/%s" % (url_header, telescope, location_id, file_)
    else:
        raise(ValueError("@Cham: This is not an option!"))
    return url


def aspcapStar_url(location_id, file_, url_header=None):

    """ aspcapStar url generator
    which in principle is able to generate file path

    Parameters
    ----------
    location_id: int
        for 'apo1m', it's 1
        for 'apo25m', it's like PLATE
    file_: string
        FILE
    url_header: string
        if None|'sas', it's set to be
        "https://data.sdss.org/sas/dr13/apogee/spectro/redux/%s/stars/l30e/l30e.2"%version

    Returns
    -------
    url: string
    the url of apStar file

    Note
    ----
    version: string
        currently it's 'r6' @20161031
    """

    if url_header is None or url_header is "sas":
        url_header = ("https://data.sdss.org/sas/dr13/apogee"
                      "/spectro/redux/r6/stars/l30e/l30e.2")

    url_header = url_header.strip()
    file_ = file_.strip()

    try:
        url = "%s/%s/%s" % (url_header, location_id, file_)
    except:
        raise (ValueError("@Cham: This is not an option!"))

    return url


def apStar_download(url, file_path, verbose=False,
                    username="sdss", password="2.5-meters"):
    """ apStar file downloading utils
    which in principle is able to download everything from a valid url

    Parameters
    ----------
    url: string
        the url of the target
    file_path: string
        the path of the file to be saved
    verbose:
        if True, print status

    Returns
    -------
    status: bool
        True if success.

    """

    # # if exists, do nothing
    # if os.path.exists(file_path):
    #     try:
    #         fits.open(file_path)
    #         return True
    #     except Exception:
    #         pass

    # credentials for sdss
    p = urllib.request.HTTPPasswordMgrWithDefaultRealm()
    p.add_password(None, url, username, password)
    handler = urllib.request.HTTPBasicAuthHandler(p)
    opener = urllib.request.build_opener(handler)
    urllib.request.install_opener(opener)

    # request data
    status = True
    try:
        local_file_path, header = urllib.request.urlretrieve(url, file_path)
    except Exception:
        status = False

    # verbose
    if verbose:
        if status:
            print("@Cham: success: {0}".format(url))
        else:
            print("@Cham: failed: {0}".format(url))

    return status


def mkdir_loop(file_path, n_loop=3, verbose=True):
    """ a weak version of os.makedirs()
    which may avoid infinite loop

    Parameters
    ----------
    file_path: string
        file path
    n_loop: int
        make n-th parent directory possible to be created
    verbose: bool
        if True, verbose

    Returns
    -------
    bool

    """
    dirname = os.path.dirname(file_path)
    if n_loop > 0:

        if os.path.exists(dirname):
            # if dirname exists
            if verbose:
                print("@Cham: [n_loop=%s] dir exists: %s " % (n_loop, dirname))
            return True
        elif os.path.exists(os.path.dirname(dirname)):
            # if dirname doesn't exist, but dirname(dirname) exists --> mkdir
            if verbose:
                print("@Cham: [n_loop=%s] mkdir: %s ..." % (n_loop, dirname))
            os.mkdir(dirname)
            return True
        else:
            # if dirname(dirname) doesn't exists
            if mkdir_loop(dirname, n_loop-1):
                return mkdir_loop(file_path, n_loop)
            else:
                return False
    else:
        if verbose:
            print("@Cham: [n_loop=%s] unable to mkdir %s ..." % (n_loop, dirname))
        return False



# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Sat Sep 03 16:00:00 2017

Modifications
-------------
- Sat Sep 03 12:00:00 2017

Aims
----
- plotting tools

"""


import numpy as np
import matplotlib.pyplot as plt


def plot_mse(s):
    fig, ax = plt.subplots(1, 1, figsize=(8, 6), tight_layout=True)
    plt.hist(-s.nmse[s.nmse != 0], np.linspace(0, 1, 80), histtype='step',
             lw=2, label="MSE")
    plt.hist(-s.scores[s.nmse != 0], np.linspace(0, 1, 80), histtype='step',
             lw=2, label="CV MSE")
    ylim = plt.gca().get_ylim()
    plt.vlines(np.percentile(-s.nmse[s.nmse != 0], [14, 50, 86]), *ylim,
               linestyle='--', label="14, 50, 86 percentiles")
    plt.xlim(0, 1)
    plt.ylim(*ylim)
    plt.ylabel("Counts")
    plt.xlabel("MSE")
    fig.tight_layout()
    return fig


# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Thu Nov 17 19:00:00 2016

Modifications
-------------
- Thu Nov 17 19:00:00 2016

Aims
----
- Lndi class:
    An interpolator making use of scipy.interpolate.LinearNDInterpolator.
    This is particularly designed for synthetic spectra.

"""

from __future__ import print_function

import os

import numpy as np
from scipy.interpolate import LinearNDInterpolator
from astropy.table import Table
from joblib import load, dump, Parallel, delayed

from .normalization import normalize_spectra_block
from .hyperparameter import summarize_hyperparameters_to_table, summarize_table
from .predict import predict_labels, predict_labels_chi2, predict_spectrum
from .standardization import standardize, standardize_ivar
from .train import train_multi_pixels, train_single_pixel
from .mcmc import predict_label_mcmc
from .diagnostic import compare_labels, single_pixel_diagnostic

__all__ = ['Lndi']


class Lndi(object):
    """
    An interpolator implemented based on scipy.interpolate.LinearNDInterpolator
    """
    wave = None
    tr_flux = None
    tr_label = None
    lndi = None
    trained = False

    def __init__(self, wave, tr_flux, tr_label):
        """ Constructor of an Lndi instance """
        self.wave = wave
        self.tr_label = tr_label
        self.tr_flux = tr_flux

        self.lndi = LinearNDInterpolator(tr_label, tr_flux)
        self.trained = True

        self.ntrain = tr_label.shape[0]
        self.ndim = tr_label.shape[1]

    def predict_spectra(self, test_label):
        """ predict spectra given labels """
        return self.lndi(test_label)

    def generate_spectra_rand(self, X_init, X_std, n_rand):
        """ generate spectra randomly

        Parameters
        ----------
        X_init: ndarray (1, ndim)
            the central X
        X_std: ndarray (1, ndim)
            the std of gaussian random numbers
        n_rand: int
            the central X

        """
        X_rand = np.random.randn(n_rand, self.ndim)
        X_rand = X_init + X_rand * X_std.reshape(1, self.ndim)
        return self.predict_spectra(X_rand)

    def generate_label_rand(self, X_init, X_std, n_rand):
        """ generate labels randomly

        Parameters
        ----------
        X_init: ndarray (1, ndim)
            the central X
        X_std: ndarray (1, ndim)
            the std of gaussian random numbers
        n_rand: int
            the central X

        """
        X_rand = np.random.randn(n_rand, self.ndim)
        X_rand = X_init + X_rand * X_std.reshape(1, self.ndim)
        return X_rand

    def predict_label_rand(self, flux_goal, X_init, X_std,
                           n_rand, frac_ext=4, n_0_th=2,
                           verbose=False):
        n_loop = 0
        n_0 = 0
        while True:
            n_loop += 1

            if verbose:
                print('@Cham: n_loop = %s ...' % n_loop)

            # generate random numbers
            n_rand = np.int(n_rand)
            frac_ext = np.int(frac_ext)
            n_rand_2 = n_rand / frac_ext
            n_rand_1 = n_rand - n_rand_2
            X_rand_1 = self.generate_label_rand(X_init, X_std, n_rand_1)
            X_rand_2 = self.generate_label_rand(X_init, X_std*3., n_rand_2)
            X_rand = np.vstack((X_init, X_rand_1, X_rand_2))
            flux_rand = self.predict_spectra(X_rand)

            # kick nan
            ind_nan = np.any(np.isnan(flux_rand), axis=1)
            X_rand = X_rand[~ind_nan]
            flux_rand = flux_rand[~ind_nan]
            # find chi2_min
            i_min, chi2_min, flux_cont_min = best_chi2(
                self.wave, flux_rand, flux_goal)

            if verbose:
                print('@Cham: ', i_min, chi2_min)

            # if n_0 > n_0_th, but stuck in the same place --> end
            if i_min == 0:
                n_0 += 1
                if n_0 > n_0_th:
                    break
            else:
                n_0 = 0
                X_init = X_rand[i_min]

        return(X_rand[i_min], chi2_min,
               flux_cont_min, flux_rand[i_min]*flux_cont_min)


def best_chi2(wave, flux, flux_goal, ivar=None):
    flux_cont = determine_continuum(wave, flux, flux_goal,
                                    norm_range=(4000, 8000),
                                    dwave=100, p=(1E-7, 1E-7))
    if ivar is None:
        chi2 = np.nansum((flux*flux_cont-flux_goal)**2., axis=1)
    else:
        chi2 = np.nansum((flux*flux_cont - flux_goal) ** 2. *
                         ivar.reshape(1, -1), axis=1)

    i_min = np.argmin(chi2)
    chi2_min = chi2[i_min]
    return i_min, chi2_min, flux_cont[i_min]


def determine_continuum(wave, flux, flux_goal,
                        norm_range=(4000, 8000), dwave=100, p=(1E-7, 1E-7),
                        q=.5, rsv_frac=3):
    """ determine the best continuum making flux to fit flux_goal """
    # TODO: ivar should be considered for emission line stars
    flux_cont = normalize_spectra_block(wave, flux_goal/flux,
                                        norm_range=norm_range, dwave=dwave,
                                        p=p, q=q, ivar_block=None,
                                        rsv_frac=rsv_frac,
                                        n_jobs=1, verbose=False)[1]
    return flux_cont



# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Mon Sep 05 12:00:00 2016

Modifications
-------------
- Mon Sep 05 12:00:00 2016

Aims
----
- SVR hyper-parameters

"""

import numpy as np
from astropy.table import Table
from sklearn.svm import SVR
from sklearn.model_selection import GridSearchCV
from copy import deepcopy
import pandas as pd
from joblib import Parallel, delayed


__all__ = ['summarize_hyperparameters_to_table', 'summarize_table']


# ############################ #
# to summarize grid parameters #
# ############################ #
def hyperparameter_grid_stats(svrs, pivot=("param_C", "param_gamma"),
                              n_jobs=10, verbose=10):
    """ statistics for GridSearchCV results """
    stats_train = []
    stats_test = []
    r = Parallel(n_jobs=n_jobs, verbose=verbose)(
        delayed(hyperparameter_grid_stats_)(svr, pivot=pivot) for svr in svrs)
    for i in range(len(r)):
        stats_train.append(r[i][0])
        stats_test.append(r[i][1])
    return stats_train, stats_test


def hyperparameter_grid_stats_(svr, pivot=("param_C", "param_gamma")):
    """ statistics for GridSearchCV results """
    if isinstance(svr, GridSearchCV):
        # yes, that's it
        cvr = svr.cv_results_
        stats_train_ = deepcopy(cvr)
        stats_test_ = deepcopy(cvr)
        for k in cvr.keys():
            if k.find("test") > -1:
                stats_train_.pop(k)
            elif k.find("train") > -1:
                stats_test_.pop(k)

        if pivot is not None:
            return (
                pd.DataFrame(stats_train_).pivot(*pivot, "mean_train_score"),
                pd.DataFrame(stats_test_).pivot(*pivot, "mean_test_score"))
        else:
            return pd.DataFrame(stats_train_), pd.DataFrame(stats_test_)
    else:
        return pd.DataFrame(), pd.DataFrame()


# ######################## #
# summarize best estimator #
# ######################## #

def summarize_hyperparameters_to_table(svrs):
    """ summarize hyper-parameters as a Table

    Parameters
    ----------
    svrs: list of sklearn.svm.SVR objects
        a list of fitted SVR objets

    """
    hyperparams = []
    for svr in svrs:
        if isinstance(svr, SVR):
            hyperparams.append((svr.C, svr.gamma, svr.epsilon))
        elif isinstance(svr, GridSearchCV):
            hyperparams.append((svr.best_estimator_.C,
                                svr.best_estimator_.gamma,
                                svr.best_estimator_.epsilon))
    hp_array = np.array(hyperparams)
    return Table(data=hp_array, names=['C', 'gamma', 'epsilon'])


def summarize_table(hpt):
    """ summarize table data

    Parameters
    ----------
    hpt: astropy.table.Table
        a table of parameter

    """

    # simply use pandas.DataFrame.describe()
    print(hpt.to_pandas().describe())

    return



# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Thu Jan 19 16:00:00 2017

Modifications
-------------
- Thu Jan 19 16:00:00 2017

Aims
----
- a customized logger for SLAM

"""

# verbose level:
# 1 debug
# 2 info
# 3 warning
# 4 error
# 5 critical


import logging

# configuration
format_slam = "[%(asctime)s] [%(module)s] %(levelname)s: %(message)s"
logging.basicConfig(format=format_slam, level=0)

# create a logger for SLAM
logger = logging.getLogger('SLAM')

# how to use it in other modules:
# from .logger import logger
# logger.info("msg")



# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Fri Sep 02 13:00:00 2016

Modifications
-------------
- Fri Sep 02 13:00:00 2016

Aims
----
- standardization

"""

import numpy as np
from copy import deepcopy

from sklearn import preprocessing


def standardize(X, weight=None, robust=False):
    """ Standardize X (flux / labels)

    Parameters
    ----------
    X: ndarray
        data array

    Returns
    -------
    scaler: sklearn.StandardScaler
        scaler

    X_scaled: ndarray
        scaled X

    """
    if weight is None:
        weight = np.ones_like(X, int)
    weight = np.logical_and(weight > 0, np.isfinite(X))

    ind_good = weight > 0
    n_good = np.sum(ind_good, axis=0)

    scaler = preprocessing.StandardScaler()
    n_col = X.shape[1]

    scaler.scale_ = np.ones((n_col,), float)
    scaler.mean_ = np.zeros((n_col,), float)

    if robust:
        # estimate using percentiles
        for i_col in range(n_col):
            if n_good[i_col] > 0:
                # at least 1 good pixels
                scaler.mean_[i_col] = (np.nanpercentile(
                    X[ind_good[:, i_col], i_col], 84) + np.nanpercentile(
                    X[ind_good[:, i_col], i_col], 16)) / 2.
                scaler.scale_[i_col] = (np.nanpercentile(
                    X[ind_good[:, i_col], i_col], 84) - np.nanpercentile(
                    X[ind_good[:, i_col], i_col], 16)) / 2.

    else:
        # estimate using mean and std
        for i_col in range(n_col):
            if n_good[i_col] > 0:
                # at least 1 good pixels
                scaler.scale_[i_col] = np.std(X[ind_good[:, i_col], i_col])
                scaler.mean_[i_col] = np.mean(X[ind_good[:, i_col], i_col])

    scaler.scale_ = np.where(scaler.scale_ < 1e-300, 1., scaler.scale_)
    scaler.robust = robust
    X_scaled = scaler.transform(X)
    return scaler, X_scaled


def standardize_ivar(ivar, flux_scaler):
    """ ivar_scaler is copied from flux_scaler, but mean_ is set to be 0
    """
    # copy flux_scaler & generate ivar_scaler
    ivar_scaler = deepcopy(flux_scaler)
    ivar_scaler.mean_ *= 0
    ivar_scaler.scale_ **= -2.  # this is extremely important!
    # transform ivar data
    ivar_scaled = ivar_scaler.transform(ivar)
    return ivar_scaler, ivar_scaled


if __name__ == '__main__':
    import numpy as np

    x = np.random.randn(10, 20)
    s, xs = standardize(x)
    print (x, xs)
    print (s, s.mean_, s.scale_)



# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Sat Sep 03 12:00:00 2016

Modifications
-------------
- Sat Sep 03 12:00:00 2016

Aims
----
- define cost functions
 - temporally only chi2 is implemented

"""

from __future__ import division

import numpy as np
from scipy.stats import chisquare


__all__ = ['chi2_simple_1d']


def chi2_simple_1d(spec_obs, spec_pred, ivar=None):
    """ Calculate ivar-weighted chi-square for two spectra

    Parameters
    ----------
    spec_obs: array-like
        observed flux

    spec_pred: array-like
        predicted flux

    ivar: array-like or None
        inverse variance / weight of pixels
        if None, assume pixels are even-weighted

    """
    if ivar is None:
        # ivar not specified
        chi2_value = np.nansum((np.array(spec_obs).flatten() - np.array(
            spec_pred).flatten()) ** 2.)
    else:
        # ivar specified
        chi2_value = np.nansum((np.array(spec_obs).flatten() - np.array(
            spec_pred).flatten()) ** 2. * np.array(ivar).flatten())
    # print('chi2: ', chi2_value)
    return chi2_value


def chi2(a, b):
    c = chisquare([1, 2, 3], [1, 1, 1])
    return c.statistic



# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Sat Sep 03 12:00:00 2016

Modifications
-------------
- Sat Sep 03 12:00:00 2016

Aims
----
- utils for training SVRs

"""

import numpy as np
from joblib import Parallel, delayed
from scipy import stats
from scipy.optimize import minimize
from sklearn import svm, model_selection
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV


def train_single_pixel(X, y, sample_weight=None, cv=10,
                       **kwargs):
    """ train a single pixel, simply CV

    Parameters
    ----------
    X: ndarray with shape (n_obs x n_dim)
        X in sklearn notation
    y: ndarray with shape (n_obs, ) --> 1D
        y in sklearn notation
    sample_weight: ndarray with shape (n_obs, ) --> 1D
        weight for sample data
    cv: int / None
        if cv>=3, Cross-Validation will be performed to calculate MSE

    kwargs:
        extra kwargs will be passed to svm.SVR() method
        e.g., C=1.0, gamma='auto', epsilon=0.1

    Returns
    -------
    svm.SVR() instance & score
    if CV is not performed, score = np.nan

    """
    # instantiate SVR
    svr = svm.SVR(**kwargs)

    if sample_weight is None:
        sample_weight = np.ones_like(y, float)
    ind_use = sample_weight > 0
    X_ = X[ind_use]
    y_ = y[ind_use]
    sample_weight_ = sample_weight[ind_use]

    # fit data
    svr.fit(X_, y_, sample_weight=sample_weight_)

    # Cross-Validation
    if cv is None or cv < 2:
        # no cross-validation will be performed
        score = -np.mean(np.square(svr.predict(X_) - y_))
    else:
        # cross-validation will be performed to calculate MSE
        assert isinstance(cv, int) and cv >= 2
        scores = model_selection.cross_val_score(
            svr, X_, y_, scoring='neg_mean_squared_error', cv=cv)
        score = scores.mean()

    return svr, score


def train_single_pixel_grid(X, y, sample_weight=None, cv=10,
                            param_grid=None, **kwargs):
    """ train a single pixel using GridSearchCV

    Parameters
    ----------
    X: ndarray with shape (n_obs x n_dim)
        X in sklearn notation
    y: ndarray with shape (n_obs, ) --> 1D
        y in sklearn notation
    sample_weight: ndarray with shape (n_obs, ) --> 1D
        weight for sample data
    cv: int / None
        if cv>=3, Cross-Validation will be performed to calculate MSE
    param_grid: dict
        key, value pairs of hyper-parameter grids
        >>> param_grid = dict(C=2. ** np.arange(-5., 6.),
        >>>                   epsilon=[0.01, 0.05, 0.1, 0.15],
        >>>                   gamma=['auto', 0.2, 0.25, 0.3, 0.5])

    kwargs:
        extra kwargs will be passed to svm.SVR() method
        e.g., C=1.0, gamma='auto', epsilon=0.1

    Returns
    -------
    svm.SVR() instance & best hyper-parameters & score
    if CV is not performed, score = np.nan

    """

    # default param_grid
    if param_grid is None:
        param_grid = dict(C=2. ** np.arange(-5., 6.),
                          epsilon=[0.01, 0.05, 0.1, 0.15],
                          gamma=['auto', 0.2, 0.25, 0.3, 0.5])
    # instantiate SVR
    svr = svm.SVR(**kwargs)
    # perform GridSearchCV
    grid = GridSearchCV(svr, param_grid, cv=cv,
                        fit_params={'sample_weight': sample_weight},
                        scoring='neg_mean_squared_error', n_jobs=1)
    # fit data
    grid.fit(X, y)

    # return (svr, score)
    return grid, grid.best_score_


def train_single_pixel_rand(X, y, sample_weight=None, cv=10,
                            n_iter=100, param_dist=None, **kwargs):
    """ train a single pixel using RandomizedSearchCV

    Parameters
    ----------
    X: ndarray with shape (n_obs x n_dim)
        X in sklearn notation
    y: ndarray with shape (n_obs, ) --> 1D
        y in sklearn notation
    sample_weight: ndarray with shape (n_obs, ) --> 1D
        weight for sample data
    cv: int / None
        if cv>=3, Cross-Validation will be performed to calculate MSE
    n_iter: int
        the number of sampling of the random subset of hyper-parameter space
    param_dist: dict
        key, value pairs of hyper-parameter grids
        >>> param_dist = dict(C=stats.expon(scale=3),
        >>>                   gamma=stats.expon(scale=.1))

    kwargs:
        extra kwargs will be passed to svm.SVR() method
        e.g., C=1.0, gamma='auto', epsilon=0.1

    Returns
    -------
    svm.SVR() instance & best hyper-parameters & score
    if CV is not performed, score = np.nan

    """

    # default param_grid
    if param_dist is None:
        param_dist = dict(C=stats.expon(scale=3),
                          gamma=stats.expon(scale=.1))
    # instantiate SVR
    svr = svm.SVR(**kwargs)
    # perform RandomizedSearchCV
    rand = RandomizedSearchCV(svr, param_dist, n_iter=n_iter, cv=cv,
                              fit_params={'sample_weight': sample_weight},
                              scoring='neg_mean_squared_error', n_jobs=1)
    # fit data
    rand.fit(X, y)

    # return (svr, score)
    return rand, rand.best_score_


def svr_mse(hyperparam, X, y, verbose=False):
    """ Cross-Validation MES for SVR """
    gamma, C, epsilon = 10. ** np.array(hyperparam)

    # instantiate
    svr = svm.SVR(gamma=gamma, C=C, epsilon=epsilon)

    # MSE
    scores = model_selection.cross_val_score(
        svr, X, y, scoring='neg_mean_squared_error', cv=10, verbose=False)
    score = -scores.mean()

    # verbose
    if verbose:
        print(gamma, C, epsilon, score)

    return score


def train_single_pixel_mini(X, y, sample_weight=None, cv=10, **kwargs):
    """ train a single pixel using minize

    Parameters
    ----------
    X: ndarray with shape (n_obs x n_dim)
        X in sklearn notation
    y: ndarray with shape (n_obs, ) --> 1D
        y in sklearn notation
    sample_weight: ndarray with shape (n_obs, ) --> 1D
        weight for sample data
    cv: int / None
        if cv>=3, Cross-Validation will be performed to calculate MSE

    kwargs:
        extra parameters that will be passed to svm.SVR()


    Returns
    -------
    svm.SVR() instance & best hyper-parameters & score
    if CV is not performed, score = np.nan

    """
    # find optimized hyper-parameters
    hp0 = (-2., .7, -.15)
    hp = minimize(svr_mse, hp0, args=(X, y, sample_weight))
    gamma, C, epsilon = 10. ** np.array(hp)

    # specify hyper-parameters directly
    return train_single_pixel(X, y, sample_weight=sample_weight, cv=cv,
                              gamma=gamma, C=C, epsilon=epsilon, **kwargs)


def train_multi_pixels(X, ys, sample_weights, cv=1,
                       method='simple', n_jobs=1, verbose=10, **kwargs):
    """ train multi pixels

    Parameters
    ----------
    X: ndarray with shape (n_obs x n_dim)
        X in sklearn notation
    ys: ndarray with shape (n_obs x n_pix) -->
        y in sklearn notation
    sample_weights: ndarray
        weight of sample data
    cv: int
        number of fold in Cross-Validation
    method: string
        {'simple', 'grid', 'rand'}
    n_jobs: int
        number of processes that will be launched by joblib
    verbose: int
        the same as joblib.Parallel() parameter verbose
    kwargs:
        extra kwargs will be passed to svm.SVR() method

    Returns
    -------
    svm.SVR() instance

    """
    # determine method
    train_funcs = {'simple': train_single_pixel,
                   'grid': train_single_pixel_grid,
                   'rand': train_single_pixel_rand}
    train_func = train_funcs[method]

    # parallel run for SVR
    data = []
    for y, sample_weight in zip(ys, sample_weights):
        this_X = np.asarray(X, float, order='C')
        this_y = np.asarray(y, float, order='C')
        this_sw = np.asarray(sample_weight, float, order='C')
        this_ind = this_sw > 0
        data.append((this_X[this_ind], this_y[this_ind], this_sw[this_ind]))

    results = Parallel(n_jobs=n_jobs, verbose=verbose)(
        delayed(train_func)(*this_data, cv=cv, **kwargs) for this_data in data)

    # return results
    return results



import os


__all__ = ['PACKAGE_PATH', 'WELCOME', 'AUTHOR', 'PACKAGE_URL']


PACKAGE_PATH = os.path.dirname(__file__)

WELCOME = """
Welcome to **TheKeenan**. Have fun with it!
"""

AUTHOR = """
Author: Bo Zhang (bozhang@nao.cas.cn)
"""

PACKAGE_URL = """
Bitbucket url: https://bitbucket.org/hypergravity/thekeenan  # currently private
Github url:    https://github.com/hypergravity/thekeenan     # in prep.
PYPI url:      https://pypi.python.org/pypi/thekeenan        # in prep.
"""



# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Sat Sep 03 12:00:00 2016

Modifications
-------------
- Sat Sep 03 12:00:00 2016

Aims
----
- normalization

"""
from __future__ import division

import numpy as np
from joblib import Parallel, delayed

from .extern.interpolate import SmoothSpline


def normalize_spectrum(wave, flux, norm_range, dwave,
                       p=(1E-6, 1E-6), q=0.5, ivar=None, eps=1e-10,
                       rsv_frac=1.):
    """ A double smooth normalization of a spectrum

    Converted from Chao Liu's normSpectrum.m
    Updated by Bo Zhang

    Parameters
    ----------
    wave: ndarray (n_pix, )
        wavelegnth array
    flux: ndarray (n_pix, )
        flux array
    norm_range: tuple
        a tuple consisting (wave_start, wave_stop)
    dwave: float
        binning width
    p: tuple of 2 ps
        smoothing parameter between 0 and 1:
        0 -> LS-straight line
        1 -> cubic spline interpolant
    q: float in range of [0, 100]
        percentile, between 0 and 1
    ivar: ndarray (n_pix, ) | None
        ivar array, default is None
    eps: float
        the ivar threshold
    rsv_frac: float
        the fraction of pixels reserved in terms of std. default is 3.

    Returns
    -------
    flux_norm: ndarray
        normalized flux
    flux_cont: ndarray
        continuum flux

    Example
    -------
    >>> flux_norm, flux_cont = normalize_spectrum(
    >>>     wave, flux, (4000., 8000.), 100., p=(1E-8, 1E-7), q=0.5,
    >>>     rsv_frac=2.0)

    """
    if ivar is not None:
        # ivar is set
        ivar = np.where(np.logical_or(wave < norm_range[0],
                                      wave > norm_range[1]), 0, ivar)
        ivar = np.where(ivar <= eps, eps, ivar)
        # mask = ivar <= eps
        var = 1. / ivar
    else:
        # default config is even weight
        var = np.ones_like(flux)

    # wave = wave[~mask]
    # flux = flux[~mask]

    # check q region
    assert 0. < q < 1.

    # n_iter = len(p)
    n_bin = np.int(np.fix(np.diff(norm_range) / dwave) + 1)
    wave1 = norm_range[0]

    # SMOOTH 1
    # print(wave.shape, flux.shape, var.shape)
    if ivar is not None:
        ind_good_init = 1. * (ivar > 0.) * (flux > 0.)
    else:
        ind_good_init = 1. * (flux > 0.)
    ind_good_init = ind_good_init.astype(np.bool)
    # print("@Cham: sum(ind_good_init)", np.sum(ind_good_init))

    flux_smoothed1 = SmoothSpline(wave[ind_good_init], flux[ind_good_init],
                                  p=p[0], var=var[ind_good_init])(wave)
    dflux = flux - flux_smoothed1

    # collecting continuum pixels --> ITERATION 1
    ind_good = np.zeros(wave.shape, dtype=np.bool)
    for i_bin in range(n_bin):
        ind_bin = np.logical_and(wave > wave1 + (i_bin - 0.5) * dwave,
                                 wave <= wave1 + (i_bin + 0.5) * dwave)
        if np.sum(ind_bin > 0):
            # median & sigma
            bin_median = np.median(dflux[ind_bin])
            bin_std = np.median(np.abs(dflux - bin_median))
            # within 1 sigma with q-percentile
            ind_good_ = ind_bin * (
                np.abs(dflux - np.nanpercentile(dflux[ind_bin], q * 100.)) < (
                rsv_frac * bin_std))
            ind_good = np.logical_or(ind_good, ind_good_)

    ind_good = np.logical_and(ind_good, ind_good_init)
    # assert there is continuum pixels
    try:
        assert np.sum(ind_good) > 0
    except AssertionError:
        Warning("@Keenan.normalize_spectrum(): unable to find continuum! ")
        ind_good = np.ones(wave.shape, dtype=np.bool)

    # SMOOTH 2
    # continuum flux
    flux_smoothed2 = SmoothSpline(
        wave[ind_good], flux[ind_good], p=p[1], var=var[ind_good])(wave)
    # normalized flux
    flux_norm = flux / flux_smoothed2

    return flux_norm, flux_smoothed2


def normalize_spectra_block(wave, flux_block, norm_range, dwave,
                            p=(1E-6, 1E-6), q=0.5, ivar_block=None, eps=1e-10,
                            rsv_frac=3., n_jobs=1, verbose=10):
    """ normalize multiple spectra using the same configuration
    This is specially designed for TheKeenan

    Parameters
    ----------
    wave: ndarray (n_pix, )
        wavelegnth array
    flux_block: ndarray (n_obs, n_pix)
        flux array
    norm_range: tuple
        a tuple consisting (wave_start, wave_stop)
    dwave: float
        binning width
    p: tuple of 2 ps
        smoothing parameter between 0 and 1:
        0 -> LS-straight line
        1 -> cubic spline interpolant
    q: float in range of [0, 100]
        percentile, between 0 and 1
    ivar_block: ndarray (n_pix, ) | None
        ivar array, default is None
    eps: float
        the ivar threshold
    rsv_frac: float
        the fraction of pixels reserved in terms of std. default is 3.
    n_jobs: int
        number of processes launched by joblib
    verbose: int / bool
        verbose level

    Returns
    -------
    flux_norm_block: ndarray
        normalized flux

    flux_cont_block: ndarray
        continuum flux

    """
    if ivar_block is None:
        ivar_block = np.ones_like(flux_block)

    if flux_block.ndim == 1:
        flux_block.reshape(1, -1)
    n_spec = flux_block.shape[0]

    results = Parallel(n_jobs=n_jobs, verbose=verbose)(
        delayed(normalize_spectrum)(
            wave, flux_block[i], norm_range, dwave, p=p, q=q,
            ivar=ivar_block[i], eps=eps, rsv_frac=rsv_frac)
        for i in range(n_spec))

    # unpack results
    flux_norm_block = []
    flux_cont_block = []
    for result in results:
        flux_norm_block.append(result[0])
        flux_cont_block.append(result[1])

    return np.array(flux_norm_block), np.array(flux_cont_block)


def get_stable_pixels(pixel_disp, wave_arm=100, frac=0.20):
    """

    Parameters
    ----------
    pixel_disp: np.ndarray
        dispersion array
    wave_arm: int
        the arm length in terms of pixels
    frac: float
        the reserved fraction, between 0.00 and 1.00

    Returns
    -------
    ind_stable

    """
    ind_stable = np.zeros_like(pixel_disp, dtype=np.bool)

    for i in range(len(ind_stable)):
        edge_l = np.max([i - wave_arm, 0])
        edge_r = np.min([i + wave_arm, len(pixel_disp)])
        if pixel_disp[i] <= \
                np.percentile(pixel_disp[edge_l:edge_r], frac * 100.):
            ind_stable[i] = True

    return ind_stable


# TODO: this is a generalized version
def normalize_spectra(wave_flux_tuple_list, norm_range, dwave,
                      p=(1E-6, 1E-6), q=50, n_jobs=1, verbose=False):
    """ normalize multiple spectra using the same configuration

    Parameters
    ----------
    wave_flux_tuple_list: list[n_obs]
        a list of (wave, flux) tuple
    norm_range: tuple
        a tuple consisting (wave_start, wave_stop)
    dwave: float
        binning width
    p: tuple of 2 ps
        smoothing parameter between 0 and 1:
        0 -> LS-straight line
        1 -> cubic spline interpolant
    q: float in range of [0, 100]
        percentile, between 0 and 1
    n_jobs: int
        number of processes launched by joblib
    verbose: int / bool
        verbose level

    Returns
    -------
    flux_norm: ndarray
        normalized flux

    """
    pass


def test_normaliza_spectra_block():
    import os

    os.chdir('/pool/projects/TheKeenan/data/TheCannonData')

    from TheCannon import apogee
    import matplotlib.pyplot as plt

    tr_ID, wl, tr_flux, tr_ivar = apogee.load_spectra("example_DR10/Data")
    tr_label = apogee.load_labels("example_DR10/reference_labels.csv")

    test_ID = tr_ID
    test_flux = tr_flux
    test_ivar = tr_ivar

    r = normalize_spectra_block(wl, tr_flux, (15200., 16900.), 30., q=0.9,
                                rsv_frac=0.5,
                                p=(1E-10, 1E-10), ivar_block=tr_ivar,
                                n_jobs=10, verbose=10)

    flux_norm, flux_cont = r
    flux_norm = np.array(flux_norm)
    flux_cont = np.array(flux_cont)
    flux_ivar = tr_ivar * flux_cont ** 2

    fig = plt.figure()
    ax = fig.add_subplot(111)
    for i in range(10, 20):
        ofst = i * 0.5
        ax.plot(wl, tr_flux[i] + ofst, 'b')
        ax.plot(wl, flux_cont[i] + ofst, 'r')
    fig.tight_layout()
    fig.savefig(
        '/pool/projects/TheKeenan/data/TheCannonData/test_norm_spec_1.pdf')


if __name__ == '__main__':
    test_normaliza_spectra_block()



# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Tue Oct 04 13:00:00 2016

Modifications
-------------
- Tue Oct 04 13:00:00 2016

Aims
----
- to implement functions for binning pixels

"""

from __future__ import division

import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import PchipInterpolator, interp1d


def interp_pchip(wave, spec, wave_interp, extrapolate=False):
    """ interpolate for single spectrum (pchip)

    Parameters
    ----------
    wave: ndarray
        wavelength array
    spec: ndarray
        spectrum array
    wave_interp: ndarray
        wavelength array to be interpolated to
    extrapolate: bool
        if True, extrapolate
        if False, return NaNs for out-of-bounds pixels

    Returns
    -------
    spec_interp: ndarray
        interpolated spectrum

    """
    P = PchipInterpolator(wave, spec, extrapolate=extrapolate)
    spec_interp = P(wave_interp)
    return spec_interp


def interp_linear(wave, spec, wave_interp, fill_value=np.nan):
    """ interpolate for single spectrum (linear)

    Parameters
    ----------
    wave: ndarray
        wavelength array
    spec: ndarray
        spectrum array
    wave_interp: ndarray
        wavelength array to be interpolated to
    fill_value: float/nan
        fill out-of-bounds pixels with fill_value

    Returns
    -------
    spec_interp: ndarray
        interpolated spectrum

    """
    I = interp1d(wave, spec, kind='linear',
                 bounds_error=False, fill_value=fill_value)
    return I(wave_interp)


def interp_cubic(wave, spec, wave_interp, fill_value=np.nan):
    """ interpolate for single spectrum (cubic)

    Parameters
    ----------
    wave: ndarray
        wavelength array
    spec: ndarray
        spectrum array
    wave_interp: ndarray
        wavelength array to be interpolated to
    fill_value: float/nan
        fill out-of-bounds pixels with fill_value

    Returns
    -------
    spec_interp: ndarray
        interpolated spectrum

    """
    I = interp1d(wave, spec, kind='cubic',
                 bounds_error=False, fill_value=fill_value)
    return I(wave_interp)


def interp_nearest(wave, spec, wave_interp, fill_value=np.nan):
    """ interpolate for single spectrum (nearest)

    Parameters
    ----------
    wave: ndarray
        wavelength array
    spec: ndarray
        spectrum array
    wave_interp: ndarray
        wavelength array to be interpolated to
    fill_value: float/nan
        fill out-of-bounds pixels with fill_value

    Returns
    -------
    spec_interp: ndarray
        interpolated spectrum

    """
    I = interp1d(wave, spec, kind='nearest',
                 bounds_error=False, fill_value=fill_value)
    return I(wave_interp)


def add_noise_normal(flux, snr):
    """ add normal random noise for flux (single spectrum)

    Parameters
    ----------
    flux: ndarray
        flux array
    snr: float
        Signal-to-Noise Ratio

    Returns
    -------
    flux: ndarray

    """
    nsr = np.random.randn(*flux.shape) / snr
    nsr = np.where((nsr < 1.) * (nsr > -1.), nsr, np.zeros_like(flux))

    return flux * (1. + nsr)


def add_noise_gpoisson(flux, k=1.0):
    """ add SCALED Poisson random noise for flux (single spectrum)

    Parameters
    ----------
    flux: ndarray
        flux array
    k: float
        k times better Poisson noise, implemented in case Poisson is too noisy
        default value is 1.

    Returns
    -------
    flux: ndarray

    """
    nsr = np.random.randn(*flux.shape) / np.sqrt(np.abs(flux)) / k
    nsr = np.where((nsr < 1.) * (nsr > -1.), nsr, np.zeros_like(flux))

    return flux * (1. + nsr)


def add_noise_poisson(flux):
    """ add Poisson random noise for flux (single/multi spectrum)

    Parameters
    ----------
    flux: ndarray
        flux array

    Returns
    -------
    flux: ndarray

    """
    return np.random.poisson(flux)


def measure_poisson_snr(flux):
    """ measure Poisson SNR  for flux

    Parameters
    ----------
    flux: ndarray 2D
        flux

    Returns
    -------
    snr_med: ndarray
        the median Poisson SNR of flux

    """
    # Poisson SNR
    snr = np.sqrt(flux)
    # median Poisson SNR
    snr_med = np.median(snr, axis=1)

    return snr_med


def shift_poisson_snr(flux, snr):
    """ shift Poisson SNR for flux

    Parameters
    ----------
    flux: ndarray 1D/2D
        flux
    snr: float
        target snr

    Returns
    -------
    flux__ : ndarray 2D
        flux with median SNR = snr

    """

    if flux.ndim == 1:
        # 1d flux
        flux = flux.reshape(1, -1)
    elif flux.ndim > 2:
        # >2d
        raise(ValueError('The number of dimensions of input flux is larger than 2!'))

    # measure poisson SNR for flux
    snr_med = measure_poisson_snr(flux)[:, None]
    # determine scale
    scale_ = (snr_med/snr) ** 2.
    # scale flux
    flux_ = flux / scale_

    if flux.ndim == 1:
        flux_ = flux_.flatten()

    return flux_


def binning_pixels(wave, flux, ivar=None, n_pixel=3):
    """

    Parameters
    ----------
    wave: ndarray
        wavelength array
    flux: ndarray
        flux array
    ivar: ndarray
        ivar array
    n_pixel: int
        number of pixels binned

    Returns
    -------
    binned_wave: ndarray
        binned wavelength array
    binned_flux:
        binned flux array
    binned_ivar:
        binned ivar array

    """
    assert n_pixel > 0

    # default ivar
    if ivar is None:
        ivar = np.ones_like(flux)

    # determine the number of binned pixels
    n_binned = np.fix(len(flux) / n_pixel)

    # initialization
    binned_wave = np.ones(n_binned)
    binned_flux = np.ones(n_binned)
    binned_ivar = np.ones(n_binned)

    # iterate for each binned pixel [wave, flux, ivar]
    for i_pix in range(n_binned):
        binned_wave[i_pix] = np.mean(
            wave[i_pix * n_pixel:(i_pix + 1) * n_pixel])
        binned_flux[i_pix] = np.mean(
            flux[i_pix * n_pixel:(i_pix + 1) * n_pixel])
        this_ivar_array = ivar[i_pix * n_pixel:(i_pix + 1) * n_pixel]
        if np.all((this_ivar_array > 0.) * np.isfinite(this_ivar_array)):
            # all pixels are good
            # ################## binning method #################### #
            # (err1**2 + err2**2 + ... + errn**2) / n**2 = errbin**2 #
            # 1/ivar1 + 1/ivar2 + ... + 1/ivarn = n**2 /ivar         #
            # --> binning n pixels with the same error               #
            # --> improves SNR by a factor of sqrt(n)                #
            # ###################################################### #
            binned_ivar[i_pix] = n_pixel ** 2. / np.sum(1. / this_ivar_array)
        else:
            # bad pixel exists
            binned_ivar[i_pix] = 0.

    return binned_wave, binned_flux, binned_ivar


def test_interpolation():
    x = np.arange(0., 10., 1.)
    y = np.sin(x)
    plt.plot(x, y, 'r')
    xx = np.arange(0., 10., 0.2)
    plt.plot(xx, interp_pchip(x, y, xx), 'b')
    plt.plot(xx, interp_linear(x, y, xx), 'g')
    plt.plot(xx, interp_cubic(x, y, xx), 'c')
    plt.plot(xx, interp_nearest(x, y, xx), 'm')


if __name__ == "__main__":
    test_interpolation()



#!/usr/bin/env python
from __future__ import absolute_import, division

import numpy as np
import scipy.signal

# import scipy.sparse.linalg  # @UnusedImport
import scipy.sparse as sparse
from numpy import ones, zeros, prod, sin, diff, pi, inf, vstack, linspace
from scipy.interpolate import BPoly, interp1d

# from wafo import polynomial as pl
from . import polynomial as pl

__all__ = [
    'PPform', 'savitzky_golay', 'savitzky_golay_piecewise', 'sgolay2d',
    'SmoothSpline', 'pchip_slopes', 'slopes', 'stineman_interp', 'Pchip',
    'StinemanInterp', 'CubicHermiteSpline']


def savitzky_golay(y, window_size, order, deriv=0):
    """Smooth (and optionally differentiate) data with a Savitzky-Golay filter.
    The Savitzky-Golay filter removes high frequency noise from data.
    It has the advantage of preserving the original shape and
    features of the signal better than other types of filtering
    approaches, such as moving averages techhniques.

    Parameters
    ----------
    y : array_like, shape (N,)
        the values of the time history of the signal.
    window_size : int
        the length of the window. Must be an odd integer number.
    order : int
        the order of the polynomial used in the filtering.
        Must be less then `window_size` - 1.
    deriv: int
        order of the derivative to compute (default = 0 means only smoothing)

    Returns
    -------
    ys : ndarray, shape (N)
        the smoothed signal (or it's n-th derivative).

    Notes
    -----
    The Savitzky-Golay is a type of low-pass filter, particularly
    suited for smoothing noisy data. The test_doctstrings idea behind this
    approach is to make for each point a least-square fit with a
    polynomial of high order over a odd-sized window centered at
    the point.

    Examples
    --------
    >>> t = np.linspace(-4, 4, 500)
    >>> noise = np.random.normal(0, 0.05, t.shape)
    >>> noise = 0.4*np.sin(100*t)
    >>> y = np.exp( -t**2 ) + noise
    >>> ysg = savitzky_golay(y, window_size=31, order=4)
    >>> np.allclose(ysg[:10],
    ... [-0.00127789, -0.02390299, -0.04444364, -0.01738837,  0.00585856,
    ...  -0.01675704, -0.03140276,  0.00010455,  0.02099063, -0.00380031])
    True

    import matplotlib.pyplot as plt
    h=plt.plot(t, y, label='Noisy signal')
    h=plt.plot(t, np.exp(-t**2), 'k', lw=1.5, label='Original signal')
    h=plt.plot(t, ysg, 'r', label='Filtered signal')
    h=plt.legend()
    plt.show()

    References
    ----------
    .. [1] A. Savitzky, M. J. E. Golay, Smoothing and Differentiation of
       Data by Simplified Least Squares Procedures. Analytical
       Chemistry, 1964, 36 (8), pp 1627-1639.
    .. [2] Numerical Recipes 3rd Edition: The Art of Scientific Computing
       W.H. Press, S.A. Teukolsky, W.T. Vetterling, B.P. Flannery
       Cambridge University Press ISBN-13: 9780521880688
    """
    try:
        window_size = np.abs(np.int(window_size))
        order = np.abs(np.int(order))
    except ValueError:
        raise ValueError("window_size and order have to be of type int")
    if window_size % 2 != 1 or window_size < 1:
        raise TypeError("window_size size must be a positive odd number")
    if window_size < order + 2:
        raise TypeError("window_size is too small for the polynomials order")
    order_range = range(order + 1)
    half_window = (window_size - 1) // 2
    # precompute coefficients
    b = np.mat([[k ** i for i in order_range]
                for k in range(-half_window, half_window + 1)])
    m = np.linalg.pinv(b).A[deriv]
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals = y[0] - np.abs(y[1:half_window + 1][::-1] - y[0])
    lastvals = y[-1] + np.abs(y[-half_window - 1:-1][::-1] - y[-1])
    y = np.concatenate((firstvals, y, lastvals))
    return np.convolve(m, y, mode='valid')


def savitzky_golay_piecewise(xvals, data, kernel=11, order=4):
    '''
    One of the most popular applications of S-G filter, apart from smoothing
    UV-VIS and IR spectra, is smoothing of curves obtained in electroanalytical
    experiments. In cyclic voltammetry, voltage (being the abcissa) changes
    like a triangle wave. And in the signal there are cusps at the turning
    points (at switching potentials) which should never be smoothed.
    In this case, Savitzky-Golay smoothing should be
    done piecewise, ie. separately on pieces monotonic in x

    Example
    -------
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> n = 1e3
    >>> x = np.linspace(0, 25, n)
    >>> y = np.round(sin(x))
    >>> sig2 = linspace(0,0.5,50)

    # As an example, this figure shows the effect of an additive noise with a
    # variance of 0.2 (original signal (black), noisy signal (red) and filtered
    # signal (blue dots)).
    >>> noise = np.sqrt(0.2)*np.random.randn(*x.shape)
    >>> noise = np.sqrt(0.2)*np.sin(1000*x)
    >>> yn = y + noise
    >>> yr = savitzky_golay_piecewise(x, yn, kernel=11, order=4)
    >>> np.allclose(yr[:10],
    ...    [-0.02708216, -0.04295155, -0.08522043, -0.13995016, -0.1908162 ,
    ...     -0.22938387, -0.26932722, -0.30614865, -0.33942134, -0.3687596 ])
    True

    h=plt.plot(x, yn, 'r', x, y, 'k', x, yr, 'b.')
    '''
    turnpoint = 0
    last = len(xvals)
    if xvals[1] > xvals[0]:  # x is increasing?
        for i in range(1, last):  # yes
            if xvals[i] < xvals[i - 1]:  # search where x starts to fall
                turnpoint = i
                break
    else:  # no, x is decreasing
        for i in range(1, last):  # search where it starts to rise
            if xvals[i] > xvals[i - 1]:
                turnpoint = i
                break
    if turnpoint == 0:  # no change in direction of x
        return savitzky_golay(data, kernel, order)
    else:
        # smooth the first piece
        firstpart = savitzky_golay(data[0:turnpoint], kernel, order)
        # recursively smooth the rest
        rest = savitzky_golay_piecewise(
            xvals[turnpoint:], data[turnpoint:], kernel, order)
        return np.concatenate((firstpart, rest))


def sgolay2d(z, window_size, order, derivative=None):
    """
    Savitsky - Golay filters can also be used to smooth two dimensional data
    affected by noise. The algorithm is exactly the same as for the one
    dimensional case, only the math is a bit more tricky. The basic algorithm
    is as follow: for each point of the two dimensional matrix extract a sub
    - matrix, centered at that point and with a size equal to an odd number
    "window_size". for this sub - matrix compute a least - square fit of a
    polynomial surface, defined as
    p(x, y) = a0 + a1 * x + a2 * y + a3 * x2 + a4 * y2 + a5 * x * y + ... .

    Note that x and y are equal to zero at the central point.
    replace the initial central point with the value computed with the fit.
    Note that because the fit coefficients are linear with respect to the data
    spacing, they can pre - computed for efficiency. Moreover, it is important
    to appropriately pad the borders of the data, with a mirror image of the
    data itself, so that the evaluation of the fit at the borders of the data
    can happen smoothly.
    Here is the code for two dimensional filtering.

    Example
    -------
    # create some sample twoD data
    >>> x = np.linspace(-3,3,100)
    >>> y = np.linspace(-3,3,100)
    >>> X, Y = np.meshgrid(x,y)
    >>> Z = np.exp( -(X**2+Y**2))

    # add noise
    >>> noise = np.random.normal( 0, 0.2, Z.shape )
    >>> noise = np.sqrt(0.2) * np.sin(100*X)*np.sin(100*Y)
    >>> Zn = Z + noise

    # filter it
    >>> Zf = sgolay2d( Zn, window_size=29, order=4)
    >>> np.allclose(Zf[:3,:5],
    ...  [[ 0.29304073,  0.29749652,  0.29007645,  0.2695685 ,  0.23541966],
    ...    [ 0.29749652,  0.29819304,  0.28766723,  0.26524542,  0.23081572],
    ...    [ 0.29007645,  0.28766723,  0.27483445,  0.25141198,  0.21769662]])
    True

    # do some plotting
    import matplotlib.pyplot as plt
    h=plt.matshow(Z)
    h=plt.matshow(Zn)
    h=plt.matshow(Zf)
    """
    # number of terms in the polynomial expression
    n_terms = (order + 1) * (order + 2) / 2.0

    if window_size % 2 == 0:
        raise ValueError('window_size must be odd')

    if window_size ** 2 < n_terms:
        raise ValueError('order is too high for the window size')

    half_size = window_size // 2

    # exponents of the polynomial.
    # p(x,y) = a0 + a1*x + a2*y + a3*x^2 + a4*y^2 + a5*x*y + ...
    # this line gives a list of two item tuple. Each tuple contains
    # the exponents of the k-th term. First element of tuple is for x
    # second element for y.
    # Ex. exps = [(0,0), (1,0), (0,1), (2,0), (1,1), (0,2), ...]
    exps = [(k - n, n) for k in range(order + 1) for n in range(k + 1)]

    # coordinates of points
    ind = np.arange(-half_size, half_size + 1, dtype=np.float)
    dx = np.repeat(ind, window_size)
    dy = np.tile(ind, [window_size, 1]).reshape(window_size ** 2, )

    # build matrix of system of equation
    A = np.empty((window_size ** 2, len(exps)))
    for i, exp in enumerate(exps):
        A[:, i] = (dx ** exp[0]) * (dy ** exp[1])

    # pad input array with appropriate values at the four borders
    new_shape = z.shape[0] + 2 * half_size, z.shape[1] + 2 * half_size
    Z = np.zeros((new_shape))
    # top band
    band = z[0, :]
    Z[:half_size, half_size:-half_size] = band - \
                                          np.abs(np.flipud(
                                              z[1:half_size + 1, :]) - band)
    # bottom band
    band = z[-1, :]
    Z[-half_size:, half_size:-half_size] = band + \
                                           np.abs(np.flipud(
                                               z[-half_size - 1:-1, :]) - band)
    # left band
    band = np.tile(z[:, 0].reshape(-1, 1), [1, half_size])
    Z[half_size:-half_size, :half_size] = band - \
                                          np.abs(np.fliplr(
                                              z[:, 1:half_size + 1]) - band)
    # right band
    band = np.tile(z[:, -1].reshape(-1, 1), [1, half_size])
    Z[half_size:-half_size, -half_size:] = band + \
                                           np.abs(np.fliplr(
                                               z[:, -half_size - 1:-1]) - band)
    # central band
    Z[half_size:-half_size, half_size:-half_size] = z

    # top left corner
    band = z[0, 0]
    Z[:half_size, :half_size] = band - \
                                np.abs(
                                    np.flipud(np.fliplr(z[1:half_size + 1,
                                                        1:half_size + 1])) - band)
    # bottom right corner
    band = z[-1, -1]
    Z[-half_size:, -half_size:] = band + \
                                  np.abs(np.flipud(np.fliplr(
                                      z[-half_size - 1:-1,
                                      -half_size - 1:-1])) -
                                         band)

    # top right corner
    band = Z[half_size, -half_size:]
    Z[:half_size, -half_size:] = band - \
                                 np.abs(
                                     np.flipud(
                                         Z[half_size + 1:2 * half_size + 1,
                                         -half_size:]) - band)
    # bottom left corner
    band = Z[-half_size:, half_size].reshape(-1, 1)
    Z[-half_size:, :half_size] = band - \
                                 np.abs(
                                     np.fliplr(Z[-half_size:,
                                               half_size + 1:2 * half_size + 1]) - band)

    # solve system and convolve
    if derivative is None:
        m = np.linalg.pinv(A)[0].reshape((window_size, -1))
        return scipy.signal.fftconvolve(Z, m, mode='valid')
    elif derivative == 'col':
        c = np.linalg.pinv(A)[1].reshape((window_size, -1))
        return scipy.signal.fftconvolve(Z, -c, mode='valid')
    elif derivative == 'row':
        r = np.linalg.pinv(A)[2].reshape((window_size, -1))
        return scipy.signal.fftconvolve(Z, -r, mode='valid')
    elif derivative == 'both':
        c = np.linalg.pinv(A)[1].reshape((window_size, -1))
        r = np.linalg.pinv(A)[2].reshape((window_size, -1))
        return (scipy.signal.fftconvolve(Z, -r, mode='valid'),
                scipy.signal.fftconvolve(Z, -c, mode='valid'))


class PPform(object):
    """The ppform of the piecewise polynomials
                    is given in terms of coefficients and breaks.
    The polynomial in the ith interval is
        x_{i} <= x < x_{i+1}

    S_i = sum(coefs[m,i]*(x-breaks[i])^(k-m), m=0..k)
    where k is the degree of the polynomial.

    Example
    -------
    >>> import matplotlib.pyplot as plt
    >>> coef = np.array([[1,1]]) # unit step function
    >>> coef = np.array([[1,1],[0,1]]) # linear from 0 to 2
    >>> coef = np.array([[1,1],[1,1],[0,2]]) # linear from 0 to 2
    >>> breaks = [0,1,2]
    >>> self = PPform(coef, breaks)
    >>> x = linspace(-1, 3, 21)
    >>> y = self(x)
    >>> np.allclose(y, [ 0.  ,  0.  ,  0.  ,  0.  ,  0.  ,  0.  ,  0.24,  0.56,
    ...    0.96, 1.44,  2.  ,  2.24,  2.56,  2.96,  3.44,  4.  ,  0.  ,  0.  ,
    ...     0.  ,  0.  ,  0.  ])
    True

    h=plt.plot(x, y)
    """

    def __init__(self, coeffs, breaks, fill=0.0, sort=False, a=None, b=None):
        if sort:
            self.breaks = np.sort(breaks)
        else:
            self.breaks = np.asarray(breaks)
        if a is None:
            a = self.breaks[0]
        if b is None:
            b = self.breaks[-1]
        self.coeffs = np.asarray(coeffs)
        self.order = self.coeffs.shape[0]
        self.fill = fill
        self.a = a
        self.b = b

    def __call__(self, xnew):
        saveshape = np.shape(xnew)
        xnew = np.ravel(xnew)
        res = np.empty_like(xnew)
        mask = (self.a <= xnew) & (xnew <= self.b)
        res[~mask] = self.fill
        xx = xnew.compress(mask)
        indxs = np.searchsorted(self.breaks[:-1], xx) - 1
        indxs = indxs.clip(0, len(self.breaks))
        pp = self.coeffs
        dx = xx - self.breaks.take(indxs)

        v = pp[0, indxs]
        for i in range(1, self.order):
            v = dx * v + pp[i, indxs]
        values = v

        res[mask] = values
        res.shape = saveshape
        return res

    def linear_extrapolate(self, output=True):
        '''
        Return 1D PPform which extrapolate linearly outside its basic interval
        '''

        max_order = 2

        if self.order <= max_order:
            if output:
                return self
            else:
                return
        breaks = self.breaks.copy()
        coefs = self.coeffs.copy()
        # pieces = len(breaks) - 1

        # Add new breaks beyond each end
        breaks2add = breaks[[0, -1]] + np.array([-1, 1])
        newbreaks = np.hstack([breaks2add[0], breaks, breaks2add[1]])

        dx = newbreaks[[0, -2]] - breaks[[0, -2]]

        dx = dx.ravel()

        # Get coefficients for the new last polynomial piece (a_n)
        # by just relocate the previous last polynomial and
        # then set all terms of order > maxOrder to zero

        a_nn = coefs[:, -1]
        dxN = dx[-1]

        a_n = pl.polyreloc(a_nn, -dxN)  # Relocate last polynomial
        # set to zero all terms of order > maxOrder
        a_n[0:self.order - max_order] = 0

        # Get the coefficients for the new first piece (a_1)
        # by first setting all terms of order > maxOrder to zero and then
        # relocate the polynomial.

        # Set to zero all terms of order > maxOrder, i.e., not using them
        a_11 = coefs[self.order - max_order::, 0]
        dx1 = dx[0]

        a_1 = pl.polyreloc(a_11, -dx1)  # Relocate first polynomial
        a_1 = np.hstack([zeros(self.order - max_order), a_1])

        newcoefs = np.hstack([a_1.reshape(-1, 1), coefs, a_n.reshape(-1, 1)])
        if output:
            return PPform(newcoefs, newbreaks, a=-inf, b=inf)
        else:
            self.coeffs = newcoefs
            self.breaks = newbreaks
            self.a = -inf
            self.b = inf

    def derivative(self):
        """
        Return first derivative of the piecewise polynomial
        """

        cof = pl.polyder(self.coeffs)
        brks = self.breaks.copy()
        return PPform(cof, brks, fill=self.fill)

    def integrate(self):
        """
        Return the indefinite integral of the piecewise polynomial
        """
        cof = pl.polyint(self.coeffs)

        pieces = len(self.breaks) - 1
        if 1 < pieces:
            # evaluate each integrated polynomial at the right endpoint of its
            # interval
            xs = diff(self.breaks[:-1, ...], axis=0)
            index = np.arange(pieces - 1)

            vv = xs * cof[0, index]
            k = self.order
            for i in range(1, k):
                vv = xs * (vv + cof[i, index])

            cof[-1] = np.hstack((0, vv)).cumsum()

        return PPform(cof, self.breaks, fill=self.fill)


# def fromspline(self, xk, cvals, order, fill=0.0):
#         N = len(xk) - 1
#         sivals = np.empty((order + 1, N), dtype=float)
#         for m in range(order, -1, -1):
#             fact = spec.gamma(m + 1)
#             res = _fitpack._bspleval(xk[:-1], xk, cvals, order, m)
#             res /= fact
#             sivals[order - m, :] = res
#         return self(sivals, xk, fill=fill)


class SmoothSpline(PPform):
    """
    Cubic Smoothing Spline.

    Parameters
    ----------
    x : array-like
        x-coordinates of data. (vector)
    y : array-like
        y-coordinates of data. (vector or matrix)
    p : real scalar
        smoothing parameter between 0 and 1:
        0 -> LS-straight line
        1 -> cubic spline interpolant
    lin_extrap : bool
        if False regular smoothing spline
        if True a smoothing spline with a constraint on the ends to
        ensure linear extrapolation outside the range of the data (default)
    var : array-like
        variance of each y(i) (default  1)

    Returns
    -------
    pp : ppform
        If xx is not given, return self-form of the spline.

    Given the approximate values

        y(i) = g(x(i))+e(i)

    of some smooth function, g, where e(i) is the error. SMOOTH tries to
    recover g from y by constructing a function, f, which  minimizes

      p * sum (Y(i) - f(X(i)))^2/d2(i)  +  (1-p) * int (f'')^2


    Example
    -------
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> x = np.linspace(0, 1, 21)
    >>> noise = 1e-1*np.random.randn(x.size)
    >>> noise = np.array(
    ...    [-0.03298601, -0.08164429, -0.06845745, -0.20718593,  0.08666282,
    ...     0.04702094,  0.08208645, -0.1017021 , -0.03031708,  0.22871709,
    ...    -0.10302486, -0.17724316, -0.05885157, -0.03875947, -0.1102984 ,
    ...    -0.05542001, -0.12717549,  0.14337697, -0.02637848, -0.10353976,
    ...    -0.0618834 ])

    >>> y = np.exp(x) + noise
    >>> pp9 = SmoothSpline(x, y, p=.9)
    >>> pp99 = SmoothSpline(x, y, p=.99, var=0.01)

    >>> y99 = pp99(x); y9 = pp9(x)
    >>> np.allclose(y9,
    ...    [ 0.8754795 ,  0.95285289,  1.03033239,  1.10803792,  1.18606854,
    ...     1.26443234,  1.34321265,  1.42258227,  1.5027733 ,  1.58394785,
    ...     1.66625727,  1.74998243,  1.8353173 ,  1.92227431,  2.01076693,
    ...     2.10064087,  2.19164551,  2.28346334,  2.37573696,  2.46825194,
    ...     2.56087699])
    True
    >>> np.allclose(y99,
    ...     [ 0.95227461,  0.97317995,  1.01159244,  1.08726908,  1.21260587,
    ...     1.31545644,  1.37829108,  1.42719649,  1.51308685,  1.59669367,
    ...     1.61486217,  1.64481078,  1.72970022,  1.83208819,  1.93312796,
    ...     2.05164767,  2.19326122,  2.34608425,  2.45023567,  2.5357288 ,
    ...     2.6357401 ])
    True


    h=plt.plot(x,y, x,pp99(x),'g', x,pp9(x),'k', x,np.exp(x),'r')

    See also
    --------
    lc2tr, dat2tr


    References
    ----------
    Carl de Boor (1978)
    'Practical Guide to Splines'
    Springer Verlag
    Uses EqXIV.6--9, self 239
    """

    def __init__(self, xx, yy, p=None, lin_extrap=True, var=1):
        coefs, brks = self._compute_coefs(xx, yy, p, var)
        super(SmoothSpline, self).__init__(coefs, brks)
        if lin_extrap:
            self.linear_extrapolate(output=False)

    def _compute_coefs(self, xx, yy, p=None, var=1):
        x, y = np.atleast_1d(xx, yy)
        x = x.ravel()
        dx = np.diff(x)
        must_sort = (dx < 0).any()
        if must_sort:
            ind = x.argsort()
            x = x[ind]
            y = y[..., ind]
            dx = np.diff(x)

        n = len(x)

        # ndy = y.ndim
        szy = y.shape

        nd = np.int(prod(szy[:-1]))  ## modified by Bo Zhang
        ny = szy[-1]

        if n < 2:
            raise ValueError('There must be >=2 data points.')
        elif (dx <= 0).any():
            raise ValueError('Two consecutive values in x can not be equal.')
        elif n != ny:
            raise ValueError('x and y must have the same length.')

        dydx = np.diff(y) / dx

        if (n == 2):  # % straight line
            coefs = np.vstack([dydx.ravel(), y[0, :]])
        else:

            dx1 = 1. / dx
            D = sparse.spdiags(var * ones(n), 0, n, n)  # The variance

            u, p = self._compute_u(p, D, dydx, dx, dx1, n)
            dx1.shape = (n - 1, -1)
            dx.shape = (n - 1, -1)
            zrs = zeros(nd)
            if p < 1:
                # faster than yi-6*(1-p)*Q*u
                Qu = D * diff(vstack([zrs, diff(vstack([zrs, u, zrs]),
                                                axis=0) * dx1, zrs]), axis=0)
                ai = (y - (6 * (1 - p) * Qu).T).T
            else:
                ai = y.reshape(n, -1)

            # The piecewise polynominals are written as
            # fi=ai+bi*(x-xi)+ci*(x-xi)^2+di*(x-xi)^3
            # where the derivatives in the knots according to Carl de Boor are:
            #    ddfi  = 6*p*[0;u] = 2*ci;
            #    dddfi = 2*diff([ci;0])./dx = 6*di;
            #    dfi   = diff(ai)./dx-(ci+di.*dx).*dx = bi;

            ci = np.vstack([zrs, 3 * p * u])
            di = (diff(vstack([ci, zrs]), axis=0) * dx1 / 3)
            bi = (diff(ai, axis=0) * dx1 - (ci + di * dx) * dx)
            ai = ai[:n - 1, ...]
            if nd > 1:
                di = di.T
                ci = ci.T
                ai = ai.T
            if not any(di):
                if not any(ci):
                    coefs = vstack([bi.ravel(), ai.ravel()])
                else:
                    coefs = vstack([ci.ravel(), bi.ravel(), ai.ravel()])
            else:
                coefs = vstack(
                    [di.ravel(), ci.ravel(), bi.ravel(), ai.ravel()])

        return coefs, x

    @staticmethod
    def _compute_u(p, D, dydx, dx, dx1, n):
        if p is None or p != 0:
            data = [dx[1:n - 1], 2 * (dx[:n - 2] + dx[1:n - 1]), dx[:n - 2]]
            R = sparse.spdiags(data, [-1, 0, 1], n - 2, n - 2)

        if p is None or p < 1:
            Q = sparse.spdiags(
                [dx1[:n - 2], -(dx1[:n - 2] + dx1[1:n - 1]), dx1[1:n - 1]],
                [0, -1, -2], n, n - 2)
            QDQ = (Q.T * D * Q)
            if p is None or p < 0:
                # Estimate p
                p = 1. / \
                    (1. + QDQ.diagonal().sum() /
                     (100. * R.diagonal().sum() ** 2))

            if p == 0:
                QQ = 6 * QDQ
            else:
                QQ = (6 * (1 - p)) * (QDQ) + p * R
        else:
            QQ = R

        # Make sure it uses symmetric matrix solver
        ddydx = diff(dydx, axis=0)
        # sp.linalg.use_solver(useUmfpack=True)
        u = 2 * sparse.linalg.spsolve((QQ + QQ.T), ddydx)  # @UndefinedVariable
        return u.reshape(n - 2, -1), p


def _edge_case(m0, d1):
    return np.where((d1 == 0) | (m0 == 0), 0.0, 1.0 / (1.0 / m0 + 1.0 / d1))


def pchip_slopes(x, y):
    # Determine the derivatives at the points y_k, d_k, by using
    #  PCHIP algorithm is:
    # We choose the derivatives at the point x_k by
    # Let m_k be the slope of the kth segment (between k and k+1)
    # If m_k=0 or m_{k-1}=0 or sgn(m_k) != sgn(m_{k-1}) then d_k == 0
    # else use weighted harmonic mean:
    #   w_1 = 2h_k + h_{k-1}, w_2 = h_k + 2h_{k-1}
    #   1/d_k = 1/(w_1 + w_2)*(w_1 / m_k + w_2 / m_{k-1})
    #   where h_k is the spacing between x_k and x_{k+1}

    hk = x[1:] - x[:-1]
    mk = (y[1:] - y[:-1]) / hk
    smk = np.sign(mk)
    condition = ((smk[1:] != smk[:-1]) | (mk[1:] == 0) | (mk[:-1] == 0))

    w1 = 2 * hk[1:] + hk[:-1]
    w2 = hk[1:] + 2 * hk[:-1]
    whmean = 1.0 / (w1 + w2) * (w1 / mk[1:] + w2 / mk[:-1])

    dk = np.zeros_like(y)
    dk[1:-1][condition] = 0.0
    dk[1:-1][~condition] = 1.0 / whmean[~condition]

    # For end-points choose d_0 so that 1/d_0 = 1/m_0 + 1/d_1 unless
    #  one of d_1 or m_0 is 0, then choose d_0 = 0

    dk[0] = _edge_case(mk[0], dk[1])
    dk[-1] = _edge_case(mk[-1], dk[-2])
    return dk


def slopes(x, y, method='parabola', tension=0, monotone=False):
    '''
    Return estimated slopes y'(x)

    Parameters
    ----------
    x, y : array-like
        array containing the x- and y-data, respectively.
        x must be sorted low to high... (no repeats) while
        y can have repeated values.
    method : string
        defining method of estimation for yp. Valid options are:
        'Catmull-Rom'  yp = (y[k+1]-y[k-1])/(x[k+1]-x[k-1])
        'Cardinal'     yp = (1-tension) * (y[k+1]-y[k-1])/(x[k+1]-x[k-1])
        'parabola'
        'secant' average secants
           yp = 0.5*((y[k+1]-y[k])/(x[k+1]-x[k]) + (y[k]-y[k-1])/(x[k]-x[k-1]))
    tension : real scalar between 0 and 1.
        tension parameter used in Cardinal method
    monotone : bool
        If True modifies yp to preserve monoticity

    Returns
    -------
    yp : ndarray
        estimated slope

    References:
    -----------
    Wikipedia:  Monotone cubic interpolation
                Cubic Hermite spline

    '''
    x = np.asarray(x, np.float_)
    y = np.asarray(y, np.float_)
    yp = np.zeros(y.shape, np.float_)

    dx = x[1:] - x[:-1]
    # Compute the slopes of the secant lines between successive points
    dydx = (y[1:] - y[:-1]) / dx

    method = method.lower()
    if method.startswith('p'):  # parabola'):
        yp[1:-1] = (dydx[:-1] * dx[1:] + dydx[1:] * dx[:-1]) / \
                   (dx[1:] + dx[:-1])
        yp[0] = 2.0 * dydx[0] - yp[1]
        yp[-1] = 2.0 * dydx[-1] - yp[-2]
    else:
        # At the endpoints - use one-sided differences
        yp[0] = dydx[0]
        yp[-1] = dydx[-1]
        if method.startswith('s'):  # secant'):
            # In the middle - use the average of the secants
            yp[1:-1] = (dydx[:-1] + dydx[1:]) / 2.0
        else:  # Cardinal or Catmull-Rom method
            yp[1:-1] = (y[2:] - y[:-2]) / (x[2:] - x[:-2])
            if method.startswith('car'):  # cardinal'):
                yp = (1 - tension) * yp

    if monotone:
        # Special case: intervals where y[k] == y[k+1]
        # Setting these slopes to zero guarantees the spline connecting
        # these points will be flat which preserves monotonicity
        ii, = (dydx == 0.0).nonzero()
        yp[ii] = 0.0
        yp[ii + 1] = 0.0

        alpha = yp[:-1] / dydx
        beta = yp[1:] / dydx
        dist = alpha ** 2 + beta ** 2
        tau = 3.0 / np.sqrt(dist)

        # To prevent overshoot or undershoot, restrict the position vector
        # (alpha, beta) to a circle of radius 3.  If (alpha**2 +  beta**2)>9,
        # then set m[k] = tau[k]alpha[k]delta[k] and
        #  m[k+1] =  tau[k]beta[b]delta[k]
        # where tau = 3/sqrt(alpha**2 + beta**2).

        # Find the indices that need adjustment
        indices_to_fix, = (dist > 9.0).nonzero()
        for ii in indices_to_fix:
            yp[ii] = tau[ii] * alpha[ii] * dydx[ii]
            yp[ii + 1] = tau[ii] * beta[ii] * dydx[ii]

    return yp


def stineman_interp(xi, x, y, yp=None):
    """
    Given data vectors *x* and *y*, the slope vector *yp* and a new
    abscissa vector *xi*, the function :func:`stineman_interp` uses
    Stineman interpolation to calculate a vector *yi* corresponding to
    *xi*.

    Here's an example that generates a coarse sine curve, then
    interpolates over a finer abscissa::

      x = linspace(0,2*pi,20);  y = sin(x); yp = cos(x)
      xi = linspace(0,2*pi,40);
      yi = stineman_interp(xi,x,y,yp);
      plot(x,y,'o',xi,yi)

    The interpolation method is described in the article A
    CONSISTENTLY WELL BEHAVED METHOD OF INTERPOLATION by Russell
    W. Stineman. The article appeared in the July 1980 issue of
    Creative Computing with a note from the editor stating that while
    they were:

      not an academic journal but once in a while something serious
      and original comes in adding that this was
      "apparently a real solution" to a well known problem.

    For *yp* = *None*, the routine automatically determines the slopes
    using the :func:`slopes` routine.

    *x* is assumed to be sorted in increasing order.

    For values ``xi[j] < x[0]`` or ``xi[j] > x[-1]``, the routine
    tries an extrapolation.  The relevance of the data obtained from
    this, of course, is questionable...

    Original implementation by Halldor Bjornsson, Icelandic
    Meteorolocial Office, March 2006 halldor at vedur.is

    Completely reworked and optimized for Python by Norbert Nemec,
    Institute of Theoretical Physics, University or Regensburg, April
    2006 Norbert.Nemec at physik.uni-regensburg.de
    """

    # Cast key variables as float.
    x = np.asarray(x, np.float_)
    y = np.asarray(y, np.float_)
    assert x.shape == y.shape
    # N = len(y)

    if yp is None:
        yp = slopes(x, y)
    else:
        yp = np.asarray(yp, np.float_)

    xi = np.asarray(xi, np.float_)
    # yi = np.zeros(xi.shape, np.float_)

    # calculate linear slopes
    dx = x[1:] - x[:-1]
    dy = y[1:] - y[:-1]
    s = dy / dx  # note length of s is N-1 so last element is #N-2

    # find the segment each xi is in
    # this line actually is the key to the efficiency of this implementation
    idx = np.searchsorted(x[1:-1], xi)

    # now we have generally: x[idx[j]] <= xi[j] <= x[idx[j]+1]
    # except at the boundaries, where it may be that xi[j] < x[0] or xi[j] >
    # x[-1]

    # the y-values that would come out from a linear interpolation:
    sidx = s.take(idx)
    xidx = x.take(idx)
    yidx = y.take(idx)
    xidxp1 = x.take(idx + 1)
    yo = yidx + sidx * (xi - xidx)

    # the difference that comes when using the slopes given in yp
    # using the yp slope of the left point
    dy1 = (yp.take(idx) - sidx) * (xi - xidx)
    # using the yp slope of the right point
    dy2 = (yp.take(idx + 1) - sidx) * (xi - xidxp1)

    dy1dy2 = dy1 * dy2
    # The following is optimized for Python. The solution actually
    # does more calculations than necessary but exploiting the power
    # of numpy, this is far more efficient than coding a loop by hand
    # in Python
    dy1mdy2 = np.where(dy1dy2, dy1 - dy2, np.inf)
    dy1pdy2 = np.where(dy1dy2, dy1 + dy2, np.inf)
    yi = yo + dy1dy2 * np.choose(
        np.array(np.sign(dy1dy2), np.int32) + 1,
        ((2 * xi - xidx - xidxp1) / ((dy1mdy2) * (xidxp1 - xidx)), 0.0,
         1 / (dy1pdy2)))
    return yi


class StinemanInterp(object):
    '''
    Returns an interpolating function
        that runs through a set of points according to the algorithm of
        Stineman (1980).

    Parameters
    ----------
    x,y : array-like
        coordinates of points defining the interpolating function.
    yp : array-like
        slopes of the interpolating function at x.
        Optional: only given if they are known, else the argument is not used.
    method : string
        method for computing the slope at the given points if the slope is not
        known. With method= "parabola" calculates the slopes from a parabola
        through every three points.

    Notes
    -----
    The interpolation method is described by Russell W. Stineman (1980)

    According to Stineman, the interpolation procedure has "the following
    properties:

    If values of the ordinates of the specified points change monotonically,
    and the slopes of the line segments joining the points change
    monotonically, then the interpolating curve and its slope will change
    monotonically. If the slopes of the line segments joining the specified
    points change monotonically, then the slopes of the interpolating curve
    will change monotonically. Suppose that the conditions in (1) or (2) are
    satisfied by a set of points, but a small change in the ordinate or slope
    at one of the points will result conditions(1) or (2) being not longer
    satisfied. Then making this small change in the ordinate or slope at a
    point will cause no more than a small change in the interpolating
    curve." The method is based on rational interpolation with specially chosen
    rational functions to satisfy the above three conditions.

    Slopes computed at the given points with the methods provided by the
    `StinemanInterp' function satisfy Stineman's requirements.
    The original method suggested by Stineman(method="scaledstineman", the
    default, and "stineman") result in lower slopes near abrupt steps or spikes
    in the point sequence, and therefore a smaller tendency for overshooting.
    The method based on a second degree polynomial(method="parabola") provides
    better approximation to smooth functions, but it results in in higher
    slopes near abrupt steps or spikes and can lead to some overshooting where
    Stineman's method does not. Both methods lead to much less tendency for
    `spurious' oscillations than traditional interplation methods based on
    polynomials, such as splines
    (see the examples section).

    Stineman states that "The complete assurance that the procedure will never
    generate `wild' points makes it attractive as a general purpose procedure".

    This interpolation method has been implemented in Matlab and R in addition
    to Python.

    Examples
    --------
    >>> import wafo.interpolate as wi
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> x = np.linspace(0,2*pi,20)
    >>> y = np.sin(x); yp = np.cos(x)
    >>> xi = np.linspace(0,2*pi,40);
    >>> yi = wi.StinemanInterp(x,y)(xi)
    >>> np.allclose(yi[:10],
    ...    [ 0.,  0.16258231,  0.31681338,  0.46390886,  0.60091421,
    ...      0.7206556 ,  0.82314953,  0.90304148,  0.96059538,  0.99241945])
    True
    >>> yi1 = wi.CubicHermiteSpline(x,y, yp)(xi)
    >>> yi2 = wi.Pchip(x,y, method='parabola')(xi)

    h=plt.subplot(211)
    h=plt.plot(x,y,'o',xi,yi,'r', xi,yi1, 'g', xi,yi1, 'b')
    h=plt.subplot(212)
    h=plt.plot(xi,np.abs(sin(xi)-yi), 'r',
               xi,  np.abs(sin(xi)-yi1), 'g',
               xi, np.abs(sin(xi)-yi2), 'b')

    References
    ----------
    Stineman, R. W. A Consistently Well Behaved Method of Interpolation.
    Creative Computing (1980), volume 6, number 7, p. 54-57.

    See Also
    --------
    slopes, Pchip
    '''

    def __init__(self, x, y, yp=None, method='parabola', monotone=False):
        if yp is None:
            yp = slopes(x, y, method, monotone=monotone)
        self.x = np.asarray(x, np.float_)
        self.y = np.asarray(y, np.float_)
        self.yp = np.asarray(yp, np.float_)

    def __call__(self, xi):
        xi = np.asarray(xi, np.float_)
        x = self.x
        y = self.y
        yp = self.yp
        # calculate linear slopes
        dx = x[1:] - x[:-1]
        dy = y[1:] - y[:-1]
        s = dy / dx  # note length of s is N-1 so last element is #N-2

        # find the segment each xi is in
        # this line actually is the key to the efficiency of this
        # implementation
        idx = np.searchsorted(x[1:-1], xi)

        # now we have generally: x[idx[j]] <= xi[j] <= x[idx[j]+1]
        # except at the boundaries, where it may be that xi[j] < x[0] or xi[j]
        # > x[-1]

        # the y-values that would come out from a linear interpolation:
        sidx = s.take(idx)
        xidx = x.take(idx)
        yidx = y.take(idx)
        xidxp1 = x.take(idx + 1)
        yo = yidx + sidx * (xi - xidx)

        # the difference that comes when using the slopes given in yp
        # using the yp slope of the left point
        dy1 = (yp.take(idx) - sidx) * (xi - xidx)
        # using the yp slope of the right point
        dy2 = (yp.take(idx + 1) - sidx) * (xi - xidxp1)

        dy1dy2 = dy1 * dy2
        # The following is optimized for Python. The solution actually
        # does more calculations than necessary but exploiting the power
        # of numpy, this is far more efficient than coding a loop by hand
        # in Python
        dy1mdy2 = np.where(dy1dy2, dy1 - dy2, np.inf)
        dy1pdy2 = np.where(dy1dy2, dy1 + dy2, np.inf)
        yi = yo + dy1dy2 * np.choose(
            np.array(np.sign(dy1dy2), np.int32) + 1,
            ((2 * xi - xidx - xidxp1) / ((dy1mdy2) * (xidxp1 - xidx)), 0.0,
             1 / (dy1pdy2)))
        return yi


class StinemanInterp2(BPoly):
    def __init__(self, x, y, yp=None, method='parabola', monotone=False):
        if yp is None:
            yp = slopes(x, y, method, monotone=monotone)
        yyp = [z for z in zip(y, yp)]
        bpoly = BPoly.from_derivatives(x, yyp)
        super(StinemanInterp2, self).__init__(bpoly.c, x)


class CubicHermiteSpline(BPoly):
    '''
    Piecewise Cubic Hermite Interpolation using Catmull-Rom
    method for computing the slopes.
    '''

    def __init__(self, x, y, yp=None, method='Catmull-Rom'):
        if yp is None:
            yp = slopes(x, y, method, monotone=False)
        yyp = [z for z in zip(y, yp)]
        bpoly = BPoly.from_derivatives(x, yyp, orders=3)
        super(CubicHermiteSpline, self).__init__(bpoly.c, x)
        # super(CubicHermiteSpline, self).__init__(x, yyp, orders=3)


class Pchip(BPoly):
    """PCHIP 1-d monotonic cubic interpolation

    Description
    -----------
    x and y are arrays of values used to approximate some function f:
       y = f(x)
    This class factory function returns a callable class whose __call__ method
    uses monotonic cubic, interpolation to find the value of new points.

    Parameters
    ----------
    x : array
        A 1D array of monotonically increasing real values.  x cannot
        include duplicate values (otherwise f is overspecified)
    y : array
        A 1-D array of real values.  y's length along the interpolation
        axis must be equal to the length of x.
    yp : array
        slopes of the interpolating function at x.
        Optional: only given if they are known, else the argument is not used.
    method : string
        method for computing the slope at the given points if the slope is not
        known. With method="parabola" calculates the slopes from a parabola
        through every three points.

    Assumes x is sorted in monotonic order (e.g. x[1] > x[0])

    Example
    -------
    >>> import wafo.interpolate as wi

    # Create a step function (will demonstrate monotonicity)
    >>> x = np.arange(7.0) - 3.0
    >>> y = np.array([-1.0, -1,-1,0,1,1,1])

    # Interpolate using monotonic piecewise Hermite cubic spline
    >>> n = 20.
    >>> xvec = np.arange(n)/10. - 1.0
    >>> yvec = wi.Pchip(x, y)(xvec)
    >>> np.allclose(yvec, [-1.   , -0.981, -0.928, -0.847, -0.744, -0.625,
    ...    -0.496, -0.363, -0.232, -0.109,  0.   ,  0.109,  0.232,  0.363,
    ...    0.496,  0.625, 0.744,  0.847,  0.928,  0.981])
    True

    # Call the Scipy cubic spline interpolator
    >>> from scipy.interpolate import interpolate
    >>> function = interpolate.interp1d(x, y, kind='cubic')
    >>> yvec1 = function(xvec)
    >>> np.allclose(yvec1, [-1.00000000e+00, -9.41911765e-01, -8.70588235e-01,
    ...        -7.87500000e-01,  -6.94117647e-01,  -5.91911765e-01,
    ...        -4.82352941e-01,  -3.66911765e-01,  -2.47058824e-01,
    ...        -1.24264706e-01,   2.49800181e-16,   1.24264706e-01,
    ...         2.47058824e-01,   3.66911765e-01,   4.82352941e-01,
    ...         5.91911765e-01,   6.94117647e-01,   7.87500000e-01,
    ...         8.70588235e-01,   9.41911765e-01])
    True


    # Non-montonic cubic Hermite spline interpolator using
    # Catmul-Rom method for computing slopes...
    >>> yvec2 = wi.CubicHermiteSpline(x,y)(xvec)
    >>> yvec3 = wi.StinemanInterp(x, y)(xvec)

    >>> np.allclose(yvec2, [-1., -0.9405, -0.864 , -0.7735, -0.672 , -0.5625,
    ...    -0.448 , -0.3315, -0.216 , -0.1045,  0.    ,  0.1045,  0.216 ,
    ...    0.3315, 0.448 ,  0.5625,  0.672 ,  0.7735,  0.864 ,  0.9405])
    True

    >>> np.allclose(yvec3, [-1. , -0.9, -0.8, -0.7, -0.6, -0.5, -0.4, -0.3,
    ... -0.2, -0.1,  0. , 0.1,  0.2,  0.3,  0.4,  0.5,  0.6,  0.7,  0.8,  0.9])
    True

    # Plot the results
    import matplotlib.pyplot as plt
    h=plt.plot(x,    y,     'ro')
    h=plt.plot(xvec, yvec,  'b')
    h=plt.plot(xvec, yvec1, 'k')
    h=plt.plot(xvec, yvec2, 'g')
    h=plt.plot(xvec, yvec3, 'm')
    h=plt.title("pchip() step function test")

    h=plt.xlabel("X")
    h=plt.ylabel("Y")
    txt = "Comparing pypchip() vs. Scipy interp1d() vs. non-monotonic CHS"
    h=plt.title(txt)
    legends = ["Data", "pypchip()", "interp1d","CHS", 'SI']
    h=plt.legend(legends, loc="upper left")
    plt.show()

    """

    def __init__(self, x, y, yp=None, method='secant'):
        if yp is None:
            yp = slopes(x, y, method=method, monotone=True)
        yyp = [z for z in zip(y, yp)]
        bpoly = BPoly.from_derivatives(x, yyp, orders=3)
        super(Pchip, self).__init__(bpoly.c, x)
        # super(Pchip, self).__init__(x, yyp, orders=3)


def interp3(x, y, z, v, xi, yi, zi, method='cubic'):
    """Interpolation on 3-D. x, y, xi, yi should be 1-D
    and z.shape == (len(x), len(y), len(z))"""
    q = (x, y, z)
    qi = (xi, yi, zi)
    for j in range(3):
        pp = interp1d(q[j], v, axis=j, kind=method)
        v = pp(qi[j])
    return v


def somefunc(x, y, z):
    return x ** 2 + y ** 2 - z ** 2 + x * y * z


def test_interp3():
    # some input data
    x = np.linspace(0, 1, 5)
    y = np.linspace(0, 2, 6)
    z = np.linspace(0, 3, 7)
    v = somefunc(x[:, None, None], y[None, :, None], z[None, None, :])

    # interpolate
    xi = np.linspace(0, 1, 45)
    yi = np.linspace(0, 2, 46)
    zi = np.linspace(0, 3, 47)
    vi = interp3(x, y, z, v, xi, yi, zi)

    import matplotlib.pyplot as plt
    X, Y = np.meshgrid(xi, yi)
    plt.figure(1)
    plt.subplot(1, 2, 1)
    plt.pcolor(X, Y, vi[:, :, 12].T)
    plt.title('interpolated')
    plt.subplot(1, 2, 2)
    plt.pcolor(X, Y, somefunc(xi[:, None], yi[None, :], zi[12]).T)
    plt.title('exact')
    plt.show('hold')


def test_smoothing_spline():
    x = linspace(0, 2 * pi + pi / 4, 20)
    y = sin(x)  # + np.random.randn(x.size)
    pp = SmoothSpline(x, y, p=1)
    x1 = linspace(-1, 2 * pi + pi / 4 + 1, 20)
    y1 = pp(x1)
    pp1 = pp.derivative()
    pp0 = pp1.integrate()
    dy1 = pp1(x1)
    y01 = pp0(x1)
    # dy = y-y1
    import matplotlib.pyplot as plt

    plt.plot(x, y, x1, y1, '.', x1, dy1, 'ro', x1, y01, 'r-')
    plt.show('hold')
    pass
    # tck = interpolate.splrep(x, y, s=len(x))


def compare_methods():
    #
    # Sine wave test
    #
    fun = np.sin
    # Create a example vector containing a sine wave.
    x = np.arange(30.0) / 10.
    y = fun(x)

    # Interpolate the data above to the grid defined by "xvec"
    xvec = np.arange(250.) / 100.

    # Initialize the interpolator slopes
    # Create the pchip slopes
    m = slopes(x, y, method='parabola', monotone=True)
    m1 = slopes(x, y, method='parabola', monotone=False)
    m2 = slopes(x, y, method='catmul', monotone=False)
    m3 = pchip_slopes(x, y)

    # Call the monotonic piece-wise Hermite cubic interpolator
    yvec = Pchip(x, y, m)(xvec)
    yvec1 = Pchip(x, y, m1)(xvec)
    yvec2 = Pchip(x, y, m2)(xvec)
    yvec3 = Pchip(x, y, m3)(xvec)

    import matplotlib.pyplot as plt

    plt.figure()
    plt.plot(x, y, 'ro', xvec, fun(xvec), 'r')
    plt.title("pchip() Sin test code")

    # Plot the interpolated points
    plt.plot(xvec, yvec, xvec, yvec1, xvec, yvec2, 'g.', xvec, yvec3)
    plt.legend(
        ['true', 'true', 'parbola_monoton', 'parabola', 'catmul', 'pchip'],
        frameon=False, loc=0)
    plt.ioff()
    plt.show()


def demo_monoticity():
    # Step function test...
    import matplotlib.pyplot as plt
    plt.figure(2)
    plt.title("pchip() step function test")
    # Create a step function (will demonstrate monotonicity)
    x = np.arange(7.0) - 3.0
    y = np.array([-1.0, -1, -1, 0, 1, 1, 1])

    # Interpolate using monotonic piecewise Hermite cubic spline
    xvec = np.arange(599.) / 100. - 3.0

    # Create the pchip slopes
    m = slopes(x, y, monotone=True)
    #    m1 = slopes(x, y, monotone=False)
    #    m2  = slopes(x,y,method='catmul',monotone=False)
    m3 = pchip_slopes(x, y)
    # Interpolate...
    yvec = Pchip(x, y, m)(xvec)

    # Call the Scipy cubic spline interpolator
    from scipy.interpolate import interpolate as ip
    function = ip.interp1d(x, y, kind='cubic')
    yvec2 = function(xvec)

    # Non-montonic cubic Hermite spline interpolator using
    # Catmul-Rom method for computing slopes...
    yvec3 = CubicHermiteSpline(x, y)(xvec)
    yvec4 = StinemanInterp(x, y)(xvec)
    yvec5 = Pchip(x, y, m3)(xvec)  # @UnusedVariable

    # Plot the results
    plt.plot(x, y, 'ro', label='Data')
    plt.plot(xvec, yvec, 'b', label='Pchip')
    plt.plot(xvec, yvec2, 'k', label='interp1d')
    plt.plot(xvec, yvec3, 'g', label='CHS')
    plt.plot(xvec, yvec4, 'm', label='Stineman')
    # plt.plot(xvec, yvec5, 'yo', label='Pchip2')
    plt.xlabel("X")
    plt.ylabel("Y")
    plt.title("Comparing Pchip() vs. Scipy interp1d() vs. non-monotonic CHS")
    #    legends = ["Data", "Pchip()", "interp1d","CHS", 'Stineman']
    plt.legend(loc="upper left", frameon=False)
    plt.ioff()
    plt.show()


def test_func():
    from scipy import interpolate
    import matplotlib.pyplot as plt
    import matplotlib
    matplotlib.interactive(False)

    coef = np.array([[1, 1], [0, 1]])  # linear from 0 to 2
    # coef = np.array([[1,1],[1,1],[0,2]]) # linear from 0 to 2
    breaks = [0, 1, 2]
    pp = PPform(coef, breaks, a=-100, b=100)
    x = linspace(-1, 3, 20)
    y = pp(x)  # @UnusedVariable

    x = linspace(0, 2 * pi + pi / 4, 20)
    y = sin(x) + np.random.randn(x.size)
    tck = interpolate.splrep(x, y, s=len(x))  # @UndefinedVariable
    xnew = linspace(0, 2 * pi, 100)
    ynew = interpolate.splev(xnew, tck, der=0)  # @UndefinedVariable
    tck0 = interpolate.splmake(  # @UndefinedVariable
        xnew, ynew, order=3, kind='smoothest', conds=None)
    pp = interpolate.ppform.fromspline(*tck0)  # @UndefinedVariable

    plt.plot(x, y, "x", xnew, ynew, xnew, sin(xnew), x, y, "b", x, pp(x), 'g')
    plt.legend(['Linear', 'Cubic Spline', 'True'])
    plt.title('Cubic-spline interpolation')
    plt.show()

    t = np.arange(0, 1.1, .1)
    x = np.sin(2 * np.pi * t)
    y = np.cos(2 * np.pi * t)
    _tck1, _u = interpolate.splprep([t, y], s=0)  # @UndefinedVariable
    tck2 = interpolate.splrep(t, y, s=len(t), task=0)  # @UndefinedVariable
    # interpolate.spl
    tck = interpolate.splmake(t, y, order=3, kind='smoothest', conds=None)
    self = interpolate.ppform.fromspline(*tck2)  # @UndefinedVariable
    plt.plot(t, self(t))
    plt.show('hold')
    pass


def test_pp():
    coef = np.array([[1, 1], [0, 0]])  # linear from 0 to 2 @UnusedVariable

    # quadratic from 0 to 1 and 1 to 2.
    coef = np.array([[1, 1], [1, 1], [0, 2]])
    dc = pl.polyder(coef, 1)
    c2 = pl.polyint(dc, 1)  # @UnusedVariable
    breaks = [0, 1, 2]
    pp = PPform(coef, breaks)
    pp(0.5)
    pp(1)
    pp(1.5)
    dpp = pp.derivative()
    import matplotlib.pyplot as plt
    x = np.linspace(-1, 3)
    plt.plot(x, pp(x), x, dpp(x), '.')
    plt.show()


def test_docstrings():
    import doctest
    print('Testing docstrings in {}'.formate(__file__))
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)


if __name__ == '__main__':
    # test_func()
    test_docstrings()
    # test_smoothing_spline()
    # compare_methods()
    # demo_monoticity()
    # test_interp3()



# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Thu Feb 16 17:00:00 2016

Modifications
-------------
- Thu Feb 16 17:00:00 2016

Aims
----
- model:

"""


from abc import ABCMeta, abstractmethod


class Model(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def train(self):
        pass

    @abstractmethod
    def predict_single_spectrum(self):
        pass

    @abstractmethod
    def predict_multi_spectra(self):
        pass







# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Sun Jun 11 12:00:00 2017

Modifications
-------------
- Sun Jun 11 12:00:00 2017

Aims
----
- utils for processing spectra

"""
import sys
from collections import OrderedDict, Set, Mapping, deque
from numbers import Number

import numpy as np


def convolve_mask(mask, kernel_size_coef=.25, kernel_size_limit=(2, 100),
                  sink_region=(200, .5)):
    """
    
    Parameters
    ----------
    mask: array like
        initial mask. True for good pixels, False for bad ones.
    kernel_size_coef:
        the kernel_size/bad_chunk_length coefficient
    kernel_size_limit: tuple
        (lower limit, upper limit)
    sink_region: tuple
        (width, threshold fraction of good pixels in this region)
        if None, pass
        
    Returns
    -------
    convolved mask

    """
    # mask: True for good, False for bad
    # 1. kernel length: at least 2 pixel at most 100 pixels
    mask0 = np.array(mask, bool)
    mask1 = np.array(np.hstack((True, mask0, True)), int)
    mask2 = np.copy(mask0)

    mask1_diff = np.diff(mask1)
    bad_chunks = np.vstack(
        (np.where(mask1_diff < 0)[0], np.where(mask1_diff > 0)[0])).T

    bad_chunks_len = np.round(np.diff(bad_chunks, axis=1) * kernel_size_coef)
    bad_chunks_len = np.where(bad_chunks_len < kernel_size_limit[0],
                              kernel_size_limit[0], bad_chunks_len)
    bad_chunks_len = np.where(bad_chunks_len > kernel_size_limit[1],
                              kernel_size_limit[1], bad_chunks_len)

    bad_chunks_convolved = np.array(
        bad_chunks_len.reshape(-1, 1) * np.array([-1, 1]) + bad_chunks, int)
    bad_chunks_convolved = np.where(
        bad_chunks_convolved < 0, 0, bad_chunks_convolved)
    bad_chunks_convolved = np.where(
        bad_chunks_convolved >= len(mask0), len(mask0), bad_chunks_convolved)

    for i_chunk in range(bad_chunks_convolved.shape[0]):
        mask2[bad_chunks_convolved[i_chunk, 0]:bad_chunks_convolved[
            i_chunk, 1]] = False

    # 2. sink_region: second round mask convolution
    if sink_region is not None:
        ind_min = 0
        ind_max = len(mask0)

        good_frac = np.zeros_like(mask0, float)
        for i in range(ind_max):
            this_start = np.max((ind_min, i - sink_region[0]))
            this_stop = np.min((ind_max, i + sink_region[0]))
            if (this_stop - this_start) < (2 * sink_region[0]):
                if this_start == ind_min:
                    this_stop = this_start + sink_region[0]
                else:
                    this_start = this_stop - sink_region[0]
            good_frac[i] = np.sum(mask0[this_start:this_stop]) / \
                           (this_stop - this_start)
        mask2 = np.where(good_frac < sink_region[1], False, mask2)

    return mask2


def uniform(tr_labels, bins, n_pick=3, ignore_out=False, digits=8):
    """ make a uniform sample --> index stored in Slam.uniform_good

    Parameters
    ----------
    bins: list of arrays
        bins in each dim
    n_pick: int
        how many to pick in each bin
    ignore_out: bool
        if True, kick stars out of bins
        if False, raise error if there is any star out of bins 
    digits: int
        digits to form string
        
    Examples
    --------
    >>> uniform(data, [np.arange(3000, 6000, 100), np.arange(-1, 5, .2),
    >>>     np.arange(-5, 1, .1)], n_pick=1, ignore_out=False)

    Returns
    -------
    index of selected sub sample

    """

    n_obs, n_dim = tr_labels.shape
    try:
        assert len(bins) == n_dim
    except AssertionError:
        print("@utils.uniform: ", len(bins), n_dim, "don't match")

    # initiate arrays
    uniform_good = np.ones((n_obs,), bool)
    uniform_ind = np.ones_like(tr_labels, int) * np.nan

    # make IDs for bins
    for i_dim in range(n_dim):
        this_bins = bins[i_dim]
        for i_bin in range(len(this_bins) - 1):
            ind = np.logical_and(
                tr_labels[:, i_dim] > this_bins[i_bin],
                tr_labels[:, i_dim] < this_bins[i_bin + 1])
            uniform_ind[ind, i_dim] = i_bin

    # check bins covering all stars
    ind_not_in_bins = np.any(
        np.logical_not(np.isfinite(uniform_ind)), axis=1)
    if np.sum(ind_not_in_bins) > 0:
        if ignore_out:
            print("@utils.uniform: These stars are out of bins and ignored")
            print("i = ", np.where(ind_not_in_bins)[0])
            uniform_good &= np.logical_not(ind_not_in_bins)
        else:
            raise (ValueError(
                "@utils.uniform: bins not wide enough to cover all stars"))

    # make ID string for bins
    fmt = "{{:0{}.0f}}".format(digits)
    uniform_str = []
    for i_obs in range(n_obs):
        str_ = ""
        for i_dim in range(n_dim):
            str_ += fmt.format(uniform_ind[i_obs, i_dim])
        uniform_str.append(str_)
    uniform_str = np.array(uniform_str)

    # unique IDs
    u_str, u_inverse, u_counts = np.unique(
        uniform_str, return_inverse=True, return_counts=True)

    # pick stars from these bins
    ind_bin_need_to_pick = np.where(u_counts > n_pick)[0]
    for _ in ind_bin_need_to_pick:
        ind_in_this_bin = np.where(u_inverse == _)[0]
        np.random.shuffle(ind_in_this_bin)
        uniform_good[ind_in_this_bin[n_pick:]] = False

    print("@utils.uniform: [{}/{}] stars chosen to make a uniform sample!"
          "".format(np.sum(uniform_good), n_obs))

    return dict(uniform_picked=uniform_good,
                uniform_unpicked=np.logical_not(uniform_good),
                uniform_ind=uniform_ind,
                uniform_str=uniform_str,
                uniform_bins=bins,
                n_pick=n_pick,
                digits=digits,
                ignore_out=ignore_out)

unit_scale_dict = dict(
    b=1.,
    kb=1024**-1,
    mb=1024**-2,
    gb=1024**-3,
)


# deprecated
def sizeof(obj, unit='mb', verbose=False, key_removed=None):

    # get scale for unit
    try:
        scale = unit_scale_dict[unit]
    except KeyError:
        print("@sizeof: unit should be in ", unit_scale_dict.keys())

    # get size
    v_dict = OrderedDict()
    for _ in dir(obj):
        v_dict[_] = getsize(obj.__getattribute__(_))

    v_dict['_total'] = np.sum([v_dict[_] for _ in v_dict.keys()])
    for k in v_dict.keys():
        v_dict[k] = np.int(scale * v_dict[k])

    v_dict['_unit'] = unit

    if verbose:
        print(v_dict)

    return v_dict


# ########################################################################### #
# adapted from https://stackoverflow.com/questions/449560/how-do-i-determine-th
# e-size-of-an-object-in-python
# ########################################################################### #

# try: # Python 2
#     zero_depth_bases = (basestring, Number, xrange, bytearray)
#     iteritems = 'iteritems'
# except NameError: # Python 3
zero_depth_bases = (str, bytes, Number, range, bytearray)
iteritems = 'items'


def getsize(obj_0):
    """Recursively iterate to sum size of object & members."""
    def inner(obj, _seen_ids = set()):
        obj_id = id(obj)
        if obj_id in _seen_ids:
            return 0
        _seen_ids.add(obj_id)
        size = sys.getsizeof(obj)
        if isinstance(obj, zero_depth_bases):
            pass # bypass remaining control flow and return
        elif isinstance(obj, (tuple, list, Set, deque)):
            size += sum(inner(i) for i in obj)
        elif isinstance(obj, Mapping) or hasattr(obj, iteritems):
            size += sum(inner(k) + inner(v) for k, v in getattr(obj, iteritems)())
        # Check for custom object instances - may subclass above too
        if hasattr(obj, '__dict__'):
            size += inner(vars(obj))
        if hasattr(obj, '__slots__'): # can have __slots__ with __dict__
            size += sum(inner(getattr(obj, s)) for s in obj.__slots__ if hasattr(obj, s))
        return size
    return inner(obj_0)




from distutils.core import setup


if __name__ == '__main__':
    setup(
        name='astroslam',
        version='0.9.3',
        author='Bo Zhang',
        author_email='bozhang@nao.cas.cn',
        description=('A forward model using SVR to estimate stellar parameters'
                     ' from spectra.'),  # short description
        license='MIT',
        # install_requires=['numpy>=1.7','scipy','matplotlib','nose'],
        url='http://github.com/hypergravity/slam',
        classifiers=[
            "Development Status :: 6 - Mature",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3.5",
            "Topic :: Scientific/Engineering :: Astronomy",
            "Topic :: Scientific/Engineering :: Physics"],
        package_dir={'slam': 'slam',
                     'extern': 'slam/extern'},
        packages=['slam',
                  'slam/extern'],
        package_data={'slam': ['data/*.csv']},
        # include_package_data=True,
        requires=['numpy',
                  'scipy',
                  'matplotlib',
                  'astropy',
                  'sklearn',
                  'joblib',
                  'pandas',
                  'emcee',
                  'lmfit',
                  'ipyparallel']
    )



# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Tue Jun 20 10:00:00 2017

Modifications
-------------
- Tue Jun 20 10:00:00 2017

Aims
----
- error analysis

"""

import numpy as np
from scipy.optimize import minimize, least_squares, curve_fit
from matplotlib import pyplot as plt
from lmfit.models import GaussianModel


# ################ #
# likelihood fit
# ################ #

def lnprior(theta):
    mu, sigma = theta
    if sigma < 0:
        return -np.inf
    else:
        return 0


def lnlike(theta, data):
    mu, sigma = theta
    return np.sum(-(data - mu) ** 2 / sigma ** 2 / 2) - \
           len(data) * np.log(sigma)


def lnpost(theta, data):
    lp = lnprior(theta)
    if not np.isfinite(lp):
        return lp
    else:
        return lnlike(theta, data) + lp


def nlnpost(*args):
    return -lnpost(*args)


def gfit_mle(theta0, data):
    return minimize(nlnpost, theta0, args=(data))


def label_diff_mle(label1, label2):
    label1 = np.array(label1)
    label2 = np.array(label2)
    assert label1.shape == label2.shape

    n_obs, n_dim = label1.shape
    bias = np.zeros((n_dim,), dtype=float)
    scatter = np.zeros((n_dim,), dtype=float)

    for i_dim in range(n_dim):
        data = label1[:, i_dim] - label2[:, i_dim]
        theta0 = np.array([np.median(data), 2 * np.std(data)])
        x = gfit_mle(theta0, data)
        if x.success:
            bias[i_dim], scatter[i_dim] = x['x']
        else:
            Warning("@GFIT: not successful [i_dim={}]!".format(i_dim))
            print("------- X [i_dim={}] -------".format(i_dim))
            print(x)
            bias[i_dim], scatter[i_dim] = x['x']

    return bias, scatter


def test_gfit_mle():
    data = np.random.randn(10000, )
    theta0 = np.array([1., 1.])
    print(gfit_mle(theta0, data))


# ################ #
# binned fit
# ################ #

def gauss1d(x, a, b, c):
    return a / (np.sqrt(2. * np.pi) * c) * np.exp(
        -(x - b) ** 2 / c ** 2 / 2)


def gauss1d_cost(theta, x, y):
    a, b, c = theta
    if a <= 0 or c <= 0:
        return -np.inf
    return gauss1d(theta, x) - y


def gfit_bin(theta0, data):
    # hist, bin_edges = np.histogram(data, bins='auto')
    bins = np.arange(np.min(data), np.max(data), np.std(data)/3)
    hist, bin_edges = np.histogram(data, bins=bins)
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2
    # figure();
    # plot(bin_centers, hist)
    # return least_squares(gauss1d_cost, theta0, args=(bin_centers, hist))
    return curve_fit(gauss1d, bin_centers, hist, p0=theta0,
                     bounds=([0, -np.inf, 0], [np.inf, np.inf, np.inf]))


def test_gfit_bin():
    data = np.random.randn(10000, )
    theta0 = np.array([1., np.median(data), 2 * np.std(data)])
    theta = gfit_bin(theta0, data)[0]
    print(theta)

    # hist, bin_edges = np.histogram(data, bins='auto')
    # bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2
    # figure();
    # plot(bin_centers, hist)
    # plot(bin_centers, gauss1d(theta, bin_centers))


def label_diff_bin(label1, label2, plot=False):
    label1 = np.array(label1)
    label2 = np.array(label2)
    assert label1.shape == label2.shape

    n_obs, n_dim = label1.shape
    amp = np.zeros((n_dim,), dtype=float)
    bias = np.zeros((n_dim,), dtype=float)
    scatter = np.zeros((n_dim,), dtype=float)

    for i_dim in range(n_dim):
        data = label1[:, i_dim] - label2[:, i_dim]
        # data = data[np.logical_and(data>np.percentile(data, 0), data<np.percentile(data, 100))]
        theta0 = np.array([len(data), np.median(data), 1 * np.std(data)])
        popt, pcov = gfit_bin(theta0, data)
        amp[i_dim], bias[i_dim], scatter[i_dim] = popt

    if plot:
        fig = plt.figure(figsize=(3*n_dim, 4))
        for i_dim in range(n_dim):
            ax = fig.add_subplot(1, n_dim, i_dim+1)
            data = label1[:, i_dim] - label2[:, i_dim]
            # data = data[np.logical_and(data > np.percentile(data, 0),
            #                            data < np.percentile(data, 100))]
            bins = np.arange(np.min(data), np.max(data), np.std(data) / 2)
            hist, bin_edges = np.histogram(data, bins=bins)
            bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2
            ax.step(bin_edges, np.hstack((hist, 0)), where='post')
            ax.plot(bin_edges, gauss1d(bin_edges, amp[i_dim], bias[i_dim], scatter[i_dim]))

    return bias, scatter


# ############################### #
# binned gaussian fit using LMFIT
# ############################### #

def gfit_bin_lmfit(data, bins='', bin_std=3, plot=False):
    if bins == 'robust':
        bins = np.arange(np.min(data), np.max(data), np.std(data)/bin_std)

    # binned statistics
    hist, bin_edges = np.histogram(data, bins=bins)
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2
    
    # fit a gaussian model to data using LMFIT
    gm = GaussianModel()
    theta_guess = gm.guess(hist, x=bin_centers)
    fr = gm.fit(hist, theta_guess, x=bin_centers, method="least_squares")
    # fr.fit_report()

    return (fr.values['amplitude'], fr.values['center'], fr.values['sigma']),fr


def label_diff_lmfit(label1, label2, bins='robust', bin_std=3, plot=False):
    """ label difference between label2 and label1(truth)
    
    Parameters
    ----------
    label1:
        truth
    label2:
        guess
    bins: str
        "auto" is recommended
    bin_std:
        binwidth = std/bin_std
    plot: bool
        if True, plot figure

    Returns
    -------

    """
    label1 = np.array(label1)
    label2 = np.array(label2)
    assert label1.shape == label2.shape

    n_obs, n_dim = label1.shape
    amp = np.zeros((n_dim,), dtype=float)
    bias = np.zeros((n_dim,), dtype=float)
    scatter = np.zeros((n_dim,), dtype=float)
    frs = np.zeros((n_dim,), dtype=object)

    for i_dim in range(n_dim):
        data = label2[:, i_dim] - label1[:, i_dim]
        theta, frs[i_dim] = \
            gfit_bin_lmfit(data, bins=bins, bin_std=bin_std, plot=False)
        amp[i_dim], bias[i_dim], scatter[i_dim] = theta
            
    if plot:
        gm = GaussianModel()
        fig = plt.figure(figsize=(3*n_dim, 4))
        for i_dim in range(n_dim):
            ax = fig.add_subplot(1, n_dim, i_dim+1)
            data = label2[:, i_dim] - label1[:, i_dim]

            # binned statistics
            if bins == 'robust':
                bins = np.arange(
                    np.min(data), np.max(data), np.std(data) / bin_std)
            hist, bin_edges = np.histogram(data, bins=bins)
            # bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2
            # bin_xx = np.linspace(bin_edges[0], bin_edges[-1], 100)
            print(bin_edges)
            ax.hist(data, bins=bin_edges, histtype='step')
            ax.plot(bin_edges, gm.eval(frs[i_dim].params, x=bin_edges))

    return bias, scatter, frs

if __name__ == "__main__":
    test_gfit_mle()
    test_gfit_bin()



# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Tue Sep 06 21:00:00 2016

Modifications
-------------
- Tue Sep 06 21:00:00 2016

Aims
----
- diagnostics of tr/test_labels
- diagnostics of fitted spectra

"""

import numpy as np
import matplotlib.pyplot as plt

from .predict import predict_pixel_for_diagnostic


__all__ = ['compare_labels', 'compare_spectra']


def compare_labels(label1, label2, labelname1='Label1', labelname2='Label2',
                   figsize=None, figpath=None, ):
    """ compare two sets of labels

    Parameters
    ----------
    label1 : ndarray (n_obs, n_dim)
        label set 1 (X)
    label2 : ndarray (n_obs, n_dim)
        label set 2 (Y)
    labelname1: string
        name of label1
    labelname2: string
        name of label2
    figsize : tuple of float
        figure size
    figpath : string
        filepath of this figure

    """

    assert label1.shape == label2.shape
    n_obs, n_dim = label1.shape

    # default figsize
    if figsize is None:
        figsize = (3.2 * n_dim, 6)

    # draw figure
    fig = plt.figure(figsize=figsize)
    for i in range(n_dim):
        x, y = label1[:, i], label2[:, i]
        xy = np.stack([x, y])
        xlim = (np.min(xy), np.max(xy))

        # diagnal plot
        ax = fig.add_subplot(2, n_dim, i + 1)
        ax.plot(x, y, 'b.')
        ax.plot(xlim, xlim, 'k--')
        ax.set_xlim(xlim)
        ax.set_ylim(xlim)
        ax.set_xlabel('%s : %s' % (labelname1, i))
        ax.set_ylabel('%s : %s' % (labelname2, i))

        # diff plot
        ax = fig.add_subplot(2, n_dim, i + 1 + n_dim)
        ax.plot(x, y - x, 'b.')
        ax.plot(xlim, [0., 0.], 'k--')
        ax.set_xlim(xlim)
        ax.set_xlabel('%s : %s' % (labelname1, i))
        ax.set_ylabel('%s : %s - %s : %s' % (labelname2, i, labelname1, i))

    fig.tight_layout()

    if figpath is not None:
        fig.savefig(figpath)

    return fig


def compare_spectra(spectra1, spectra2=None, ofst_step=0.2, wave=None,
                    mediannorm=False, figsize=(10, 6), plt_max=100):
    """ compare one/two spectra set """
    n_spec = spectra1.shape[0]

    # if mediannorm is a float, scale spectra to median*
    if isinstance(mediannorm, float):
        for i in range(n_spec):
            spectra1[i] /= np.nanmedian(spectra1[i]) * mediannorm
        if spectra2 is not None:
            for i in range(n_spec):
                spectra2[i] /= np.nanmedian(spectra2[i]) * mediannorm

    # plot the figure
    fig = plt.figure(figsize=figsize)
    fig.add_subplot(111)
    if wave is None:
        for i in range(n_spec):
            ofst = i * ofst_step
            plt.plot(spectra1[i] + ofst, 'b')
            if spectra2 is not None:
                plt.plot(spectra2[i] + ofst, 'r')
    else:
        for i in range(n_spec):
            ofst = i * ofst_step
            plt.plot(wave, spectra1[i] + ofst, 'b')
            if spectra2 is not None:
                plt.plot(wave, spectra2[i] + ofst, 'r')

    return fig


def single_pixel_diagnostic(svrs,
                            i_pixel,
                            test_labels,
                            diag_dim=(0,),
                            labels_scaler=None,
                            flux_scaler=None):
    """ diagnostic a single pixel in 1D/2D

    Parameters
    ----------
    svrs: list of sklearn.smv.SVR instance
        k.svrs
    i_pixel: int
        No. of pixel that will be in diagnostic
    test_labels: ndarray ( n, ndim )
        test labels
    diag_dim: tuple/list
        diagnostic dimensions, e.g., (0, 1)
    labels_scaler:
        scaler for labels, e.g., k.tr_labels_scaler
    flux_scaler:
        scaler for flux, e.g., k.tr_flux_scaler

    Returns
    -------
    [X, (Y,) flux]

    """
    # assertions
    assert 1 <= len(diag_dim) <= 2
    for dim_ in diag_dim:
        assert 0 <= dim_ <= test_labels.shape[1]

    # draw scaling parameters for this pixel
    if flux_scaler is None:
        flux_mean_ = 0.
        flux_scale_ = 1.
    else:
        flux_mean_ = flux_scaler.mean_[i_pixel]
        flux_scale_ = flux_scaler.scale_[i_pixel]

    # prdict flux for this pixel
    pixel_flux = predict_pixel_for_diagnostic(
        svrs[i_pixel], test_labels,
        labels_scaler=labels_scaler,
        flux_mean_=flux_mean_,
        flux_scale_=flux_scale_)

    result = []
    for dim_ in diag_dim:
        result.append(test_labels[:, dim_])
    result.append(pixel_flux.flatten())

    # return [X, (Y,) flux]
    return result
